# Copyright (c) 2014 Montavista Software, LLC.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import copy

import ceilometerclient.client as cc
from oslo_log import log as logging
import six

from congress.datasources import datasource_driver
from congress.datasources import datasource_utils as ds_utils

LOG = logging.getLogger(__name__)


def d6service(name, keys, inbox, datapath, args):
    """Create a dataservice instance.

    This method is called by d6cage to create a dataservice
    instance.  There are a couple of parameters we found useful
    to add to that call, so we included them here instead of
    modifying d6cage (and all the d6cage.createservice calls).
    """
    return CeilometerDriver(name, keys, inbox, datapath, args)


# TODO(thinrichs): figure out how to move even more of this boilerplate
#   into DataSourceDriver.  E.g. change all the classes to Driver instead of
#   NeutronDriver, CeilometerDriver, etc. and move the d6instantiate function
#   to DataSourceDriver.
class CeilometerDriver(datasource_driver.PollingDataSourceDriver,
                       datasource_driver.ExecutionDriver):
    METERS = "meters"
    ALARMS = "alarms"
    EVENTS = "events"
    EVENT_TRAITS = "events.traits"
    ALARM_THRESHOLD_RULE = "alarms.threshold_rule"
    STATISTICS = "statistics"

    value_trans = {'translation-type': 'VALUE'}

    meters_translator = {
        'translation-type': 'HDICT',
        'table-name': METERS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'meter_id', 'translator': value_trans},
             {'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'type', 'translator': value_trans},
             {'fieldname': 'unit', 'translator': value_trans},
             {'fieldname': 'source', 'translator': value_trans},
             {'fieldname': 'resource_id', 'translator': value_trans},
             {'fieldname': 'user_id', 'translator': value_trans},
             {'fieldname': 'project_id', 'translator': value_trans})}

    alarms_translator = {
        'translation-type': 'HDICT',
        'table-name': ALARMS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'alarm_id', 'translator': value_trans},
             {'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'state', 'translator': value_trans},
             {'fieldname': 'enabled', 'translator': value_trans},
             {'fieldname': 'threshold_rule', 'col': 'threshold_rule_id',
              'translator': {'translation-type': 'VDICT',
                             'table-name': ALARM_THRESHOLD_RULE,
                             'id-col': 'threshold_rule_id',
                             'key-col': 'key', 'val-col': 'value',
                             'translator': value_trans}},
             {'fieldname': 'type', 'translator': value_trans},
             {'fieldname': 'description', 'translator': value_trans},
             {'fieldname': 'time_constraints', 'translator': value_trans},
             {'fieldname': 'user_id', 'translator': value_trans},
             {'fieldname': 'project_id', 'translator': value_trans},
             {'fieldname': 'alarm_actions', 'translator': value_trans},
             {'fieldname': 'ok_actions', 'translator': value_trans},
             {'fieldname': 'insufficient_data_actions', 'translator':
             value_trans},
             {'fieldname': 'repeat_actions', 'translator': value_trans},
             {'fieldname': 'timestamp', 'translator': value_trans},
             {'fieldname': 'state_timestamp', 'translator': value_trans},
             )}

    events_translator = {
        'translation-type': 'HDICT',
        'table-name': EVENTS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'message_id', 'translator': value_trans},
             {'fieldname': 'event_type', 'translator': value_trans},
             {'fieldname': 'generated', 'translator': value_trans},
             {'fieldname': 'traits',
              'translator': {'translation-type': 'HDICT',
                             'table-name': EVENT_TRAITS,
                             'selector-type': 'DICT_SELECTOR',
                             'in-list': True,
                             'parent-key': 'message_id',
                             'parent-col-name': 'event_message_id',
                             'field-translators':
                                 ({'fieldname': 'name',
                                   'translator': value_trans},
                                  {'fieldname': 'type',
                                   'translator': value_trans},
                                  {'fieldname': 'value',
                                   'translator': value_trans}
                                  )}}
             )}

    def safe_id(x):
        if isinstance(x, six.string_types):
            return x
        try:
            return x['resource_id']
        except KeyError:
            return str(x)

    statistics_translator = {
        'translation-type': 'HDICT',
        'table-name': STATISTICS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'meter_name', 'translator': value_trans},
             {'fieldname': 'groupby', 'col': 'resource_id',
              'translator': {'translation-type': 'VALUE',
                             'extract-fn': safe_id}},
             {'fieldname': 'avg', 'translator': value_trans},
             {'fieldname': 'count', 'translator': value_trans},
             {'fieldname': 'duration', 'translator': value_trans},
             {'fieldname': 'duration_start', 'translator': value_trans},
             {'fieldname': 'duration_end', 'translator': value_trans},
             {'fieldname': 'max', 'translator': value_trans},
             {'fieldname': 'min', 'translator': value_trans},
             {'fieldname': 'period', 'translator': value_trans},
             {'fieldname': 'period_end', 'translator': value_trans},
             {'fieldname': 'period_start', 'translator': value_trans},
             {'fieldname': 'sum', 'translator': value_trans},
             {'fieldname': 'unit', 'translator': value_trans})}

    TRANSLATORS = [meters_translator, alarms_translator, events_translator,
                   statistics_translator]

    def __init__(self, name='', keys='', inbox=None, datapath=None, args=None):
        super(CeilometerDriver, self).__init__(name, keys, inbox,
                                               datapath, args)
        datasource_driver.ExecutionDriver.__init__(self)
        self.creds = self.get_ceilometer_credentials_v2(args)
        self.ceilometer_client = cc.get_client(**self.creds)
        self.add_executable_client_methods(self.ceilometer_client,
                                           'ceilometerclient.v2.')
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'ceilometer'
        result['description'] = ('Datasource driver that interfaces with '
                                 'ceilometer.')
        result['config'] = ds_utils.get_openstack_required_config()
        result['secret'] = ['password']
        return result

    def update_from_datasource(self):
        """Read Data from Ceilometer datasource.

        And to fill up the current state of the policy engine.
        """
        LOG.debug("Ceilometer grabbing meters")
        meters = self.ceilometer_client.meters.list()
        self._translate_meters(meters)
        LOG.debug("METERS: %s" % str(self.state[self.METERS]))

        # TODO(ramineni): Ceilometer alarms is moved to separate
        # project Aodh. It's not fully functional yet.
        # Enable it back when its fully functional.

        # LOG.debug("Ceilometer grabbing alarms")
        # alarms = self.ceilometer_client.alarms.list()
        # self._translate_alarms(alarms)
        # LOG.debug("ALARMS: %s" % str(self.state[self.ALARMS]))
        # LOG.debug("THRESHOLD: %s"
        #          % str(self.state[self.ALARM_THRESHOLD_RULE]))

        LOG.debug("Ceilometer grabbing events")
        events = self.ceilometer_client.events.list()
        self._translate_events(events)
        LOG.debug("EVENTS: %s" % str(self.state[self.EVENTS]))
        LOG.debug("TRAITS: %s" % str(self.state[self.EVENT_TRAITS]))

        LOG.debug("Ceilometer grabbing statistics")
        statistics = self._get_statistics(meters)
        self._translate_statistics(statistics)
        LOG.debug("STATISTICS: %s" % str(self.state[self.STATISTICS]))

    def _get_statistics(self, meters):
        statistics = []
        names = set()
        for m in meters:
            LOG.debug("Adding meter %s" % m.name)
            names.add(m.name)
        for meter_name in names:
            LOG.debug("Getting all Resource ID for meter: %s"
                      % meter_name)
            stat_list = self.ceilometer_client.statistics.list(
                meter_name, groupby=['resource_id'])
            LOG.debug("Statistics List: %s" % stat_list)
            if (stat_list):
                for temp in stat_list:
                    temp_dict = copy.copy(temp.to_dict())
                    temp_dict['meter_name'] = meter_name
                    statistics.append(temp_dict)
        return statistics

    def get_ceilometer_credentials_v2(self, creds):
        d = {}
        d['version'] = '2'
        d['username'] = creds['username']
        d['password'] = creds['password']
        d['auth_url'] = creds['auth_url']
        d['tenant_name'] = creds['tenant_name']
        return d

    @ds_utils.update_state_on_changed(METERS)
    def _translate_meters(self, obj):
        """Translate the meters represented by OBJ into tables."""
        meters = [o.to_dict() for o in obj]

        LOG.debug("METERS: %s" % str(meters))

        row_data = CeilometerDriver.convert_objs(meters,
                                                 self.meters_translator)
        return row_data

    @ds_utils.update_state_on_changed(ALARMS)
    def _translate_alarms(self, obj):
        """Translate the alarms represented by OBJ into tables."""
        alarms = [o.to_dict() for o in obj]
        LOG.debug("ALARMS: %s" % str(alarms))

        row_data = CeilometerDriver.convert_objs(alarms,
                                                 self.alarms_translator)
        return row_data

    @ds_utils.update_state_on_changed(EVENTS)
    def _translate_events(self, obj):
        """Translate the events represented by OBJ into tables."""
        events = [o.to_dict() for o in obj]
        LOG.debug("EVENTS: %s" % str(events))

        row_data = CeilometerDriver.convert_objs(events,
                                                 self.events_translator)
        return row_data

    @ds_utils.update_state_on_changed(STATISTICS)
    def _translate_statistics(self, obj):
        """Translate the statistics represented by OBJ into tables."""
        LOG.debug("STATISTICS: %s" % str(obj))

        row_data = CeilometerDriver.convert_objs(obj,
                                                 self.statistics_translator)
        return row_data

    def execute(self, action, action_args):
        """Overwrite ExecutionDriver.execute()."""
        # action can be written as a method or an API call.
        func = getattr(self, action, None)
        if func and self.is_executable(func):
            func(action_args)
        else:
            self._execute_api(self.ceilometer_client, action, action_args)



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2010 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utility methods for working with WSGI servers."""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import sys

from oslo_log import log as logging
import routes.middleware
import webob.dec
import webob.exc

LOG = logging.getLogger(__name__)


class Request(webob.Request):
    pass


class Application(object):
    """Base WSGI application wrapper. Subclasses need to implement __call__."""

    @classmethod
    def factory(cls, global_config, **local_config):
        """Used for paste app factories in paste.deploy config files.

        Any local configuration (that is, values under the [app:APPNAME]
        section of the paste config) will be passed into the `__init__` method
        as kwargs.

        A hypothetical configuration would look like:

            [app:wadl]
            latest_version = 1.3
            paste.app_factory = nova.api.fancy_api:Wadl.factory

        which would result in a call to the `Wadl` class as

            import nova.api.fancy_api
            fancy_api.Wadl(latest_version='1.3')

        You could of course re-implement the `factory` method in subclasses,
        but using the kwarg passing it shouldn't be necessary.

        """
        return cls(**local_config)

    def __call__(self, environ, start_response):
        r"""Subclasses will probably want to implement __call__ like this:

        @webob.dec.wsgify(RequestClass=Request)
        def __call__(self, req):
          # Any of the following objects work as responses:

          # Option 1: simple string
          res = 'message\n'

          # Option 2: a nicely formatted HTTP exception page
          res = exc.HTTPForbidden(explanation='Nice try')

          # Option 3: a webob Response object (in case you need to play with
          # headers, or you want to be treated like an iterable, or or or)
          res = Response();
          res.app_iter = open('somefile')

          # Option 4: any wsgi app to be run next
          res = self.application

          # Option 5: you can get a Response object for a wsgi app, too, to
          # play with headers etc
          res = req.get_response(self.application)

          # You can then just return your response...
          return res
          # ... or set req.response and return None.
          req.response = res

        See the end of http://pythonpaste.org/webob/modules/dec.html
        for more info.

        """
        raise NotImplementedError(_('You must implement __call__'))


class Middleware(Application):
    """Base WSGI middleware.

    These classes require an application to be
    initialized that will be called next.  By default the middleware will
    simply call its wrapped app, or you can override __call__ to customize its
    behavior.

    """

    @classmethod
    def factory(cls, global_config, **local_config):
        """Used for paste app factories in paste.deploy config files.

        Any local configuration (that is, values under the [filter:APPNAME]
        section of the paste config) will be passed into the `__init__` method
        as kwargs.

        A hypothetical configuration would look like:

            [filter:analytics]
            redis_host = 127.0.0.1
            paste.filter_factory = nova.api.analytics:Analytics.factory

        which would result in a call to the `Analytics` class as

            import nova.api.analytics
            analytics.Analytics(app_from_paste, redis_host='127.0.0.1')

        You could of course re-implement the `factory` method in subclasses,
        but using the kwarg passing it shouldn't be necessary.

        """
        def _factory(app):
            return cls(app, **local_config)
        return _factory

    def __init__(self, application):
        self.application = application

    def process_request(self, req):
        """Called on each request.

        If this returns None, the next application down the stack will be
        executed. If it returns a response then that response will be returned
        and execution will stop here.

        """
        return None

    def process_response(self, response):
        """Do whatever you'd like to the response."""
        return response

    @webob.dec.wsgify(RequestClass=Request)
    def __call__(self, req):
        response = self.process_request(req)
        if response:
            return response
        response = req.get_response(self.application)
        return self.process_response(response)


class Debug(Middleware):
    """Helper class for debugging a WSGI application.

    Can be inserted into any WSGI application chain to get information
    about the request and response.

    """

    @webob.dec.wsgify(RequestClass=Request)
    def __call__(self, req):
        print(('*' * 40) + ' REQUEST ENVIRON')
        for key, value in req.environ.items():
            print(key, '=', value)
        print()
        resp = req.get_response(self.application)

        print(('*' * 40) + ' RESPONSE HEADERS')
        for (key, value) in resp.headers.items():
            print(key, '=', value)
        print()

        resp.app_iter = self.print_generator(resp.app_iter)

        return resp

    @staticmethod
    def print_generator(app_iter):
        """Iterator that prints the contents of a wrapper string."""
        print(('*' * 40) + ' BODY')
        for part in app_iter:
            sys.stdout.write(part)
            sys.stdout.flush()
            yield part
        print()


class Router(object):
    """WSGI middleware that maps incoming requests to WSGI apps."""

    def __init__(self, mapper):
        """Create a router for the given routes.Mapper.

        Each route in `mapper` must specify a 'controller', which is a
        WSGI app to call.  You'll probably want to specify an 'action' as
        well and have your controller be an object that can route
        the request to the action-specific method.

        Examples:
          mapper = routes.Mapper()
          sc = ServerController()

          # Explicit mapping of one route to a controller+action
          mapper.connect(None, '/svrlist', controller=sc, action='list')

          # Actions are all implicitly defined
          mapper.resource('server', 'servers', controller=sc)

          # Pointing to an arbitrary WSGI app.  You can specify the
          # {path_info:.*} parameter so the target app can be handed just that
          # section of the URL.
          mapper.connect(None, '/v1.0/{path_info:.*}', controller=BlogApp())

        """
        self.map = mapper
        self._router = routes.middleware.RoutesMiddleware(self._dispatch,
                                                          self.map)

    @webob.dec.wsgify(RequestClass=Request)
    def __call__(self, req):
        """Route the incoming request to a controller based on self.map.

        If no match, return a 404.

        """
        return self._router

    @staticmethod
    @webob.dec.wsgify(RequestClass=Request)
    def _dispatch(req):
        """Dispatch the request to the appropriate controller.

        Called by self._router after matching the incoming request to a route
        and putting the information into req.environ.  Either returns 404
        or the routed WSGI app's response.

        """
        match = req.environ['wsgiorg.routing_args'][1]
        if not match:
            return webob.exc.HTTPNotFound()
        app = match['controller']
        return app



# Copyright (c) 2015 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from congress.dse import deepsix


class PolicyEngineDriver(deepsix.deepSix):
    """Driver for policy engines, analogous to DataSourceDriver."""

    def set_policy(self, policy):
        """Set the policy of this policy engine. POLICY is a datalog string."""
        return NotImplementedError

    def supported_language(self):
        """Return description of the language supported by this engine.

        A description is a list of permitted tables and a list of forbidden
        tables.  Eventually we may broaden the description.
        """
        return NotImplementedError



# Copyright (c) 2013,2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import functools
import inspect
import re

from congress.datasources import constants


def get_openstack_required_config():
    return {'auth_url': constants.REQUIRED,
            'endpoint': constants.OPTIONAL,
            'region': constants.OPTIONAL,
            'username': constants.REQUIRED,
            'password': constants.REQUIRED,
            'tenant_name': constants.REQUIRED,
            'project_name': constants.OPTIONAL,
            'poll_time': constants.OPTIONAL}


def update_state_on_changed(root_table_name):
    """Decorator to check raw data before retranslating.

    If raw data is same with cached self.raw_state,
    don't translate data, return empty list directly.
    If raw data is changed, translate it and update state.
    """

    def outer(f):
        @functools.wraps(f)
        def inner(self, raw_data, *args, **kw):
            if (root_table_name not in self.raw_state or
                    # TODO(RuiChen): workaround for oslo-incubator bug/1499369,
                    # enable self.raw_state cache, once the bug is resolved.
                    raw_data is not self.raw_state[root_table_name]):
                result = f(self, raw_data, *args, **kw)
                self._update_state(root_table_name, result)
                self.raw_state[root_table_name] = raw_data
            else:
                result = []
            return result
        return inner
    return outer


def add_column(colname, desc=None):
    """Adds column in the form of dict."""
    return {'name': colname, 'desc': desc}


def inspect_methods(client, api_prefix):
    """Inspect all callable methods from client for congress."""

    # some methods are referred multiple times, we should
    # save them here to avoid infinite loop
    obj_checked = []
    method_checked = []
    # For depth-first search
    obj_stack = []
    # save all inspected methods that will be returned
    allmethods = []

    obj_checked.append(client)
    obj_stack.append(client)
    while len(obj_stack) > 0:
        cur_obj = obj_stack.pop()
        # everything starts with '_' are considered as internal only
        for f in [f for f in dir(cur_obj) if not f.startswith('_')]:
            p = getattr(cur_obj, f)
            if inspect.ismethod(p):
                m_p = {}
                # to get a name that can be called by Congress, no need
                # to return the full path
                m_p['name'] = cur_obj.__module__.replace(api_prefix, '')
                if m_p['name'] == '':
                    m_p['name'] = p.__name__
                else:
                    m_p['name'] = m_p['name'] + '.' + p.__name__
                # skip checked methods
                if m_p['name'] in method_checked:
                    continue
                m_doc = inspect.getdoc(p)
                # not return deprecated methods
                if m_doc and "DEPRECATED:" in m_doc:
                    continue

                if m_doc:
                    m_doc = re.sub('\n|\s+', ' ', m_doc)
                    x = re.split(' :param ', m_doc)
                    m_p['desc'] = x.pop(0)
                    y = inspect.getargspec(p)
                    m_p['args'] = []
                    while len(y.args) > 0:
                        m_p_name = y.args.pop(0)
                        if m_p_name == 'self':
                            continue
                        if len(x) > 0:
                            m_p_desc = x.pop(0)
                        else:
                            m_p_desc = "None"
                        m_p['args'].append({'name': m_p_name,
                                            'desc': m_p_desc})
                else:
                    m_p['args'] = []
                    m_p['desc'] = ''
                allmethods.append(m_p)
                method_checked.append(m_p['name'])
            elif inspect.isfunction(p):
                m_p = {}
                m_p['name'] = cur_obj.__module__.replace(api_prefix, '')
                if m_p['name'] == '':
                    m_p['name'] = f
                else:
                    m_p['name'] = m_p['name'] + '.' + f
                # TODO(zhenzanz): Never see doc for function yet.
                # m_doc = inspect.getdoc(p)
                m_p['args'] = []
                m_p['desc'] = ''
                allmethods.append(m_p)
                method_checked.append(m_p['name'])
            elif isinstance(p, object) and hasattr(p, '__module__'):
                # avoid infinite loop by checking that p not in obj_checked.
                # don't use 'in' since that uses ==, and some clients err
                if ((not any(p is x for x in obj_checked)) and
                   (not inspect.isbuiltin(p))):
                    if re.match(api_prefix, p.__module__):
                        if (not inspect.isclass(p)):
                            obj_stack.append(p)

    return allmethods



# Copyright 2015 VMware.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from openstack_dashboard.api import congress


LOG = logging.getLogger(__name__)


def _get_policy_tables(request):
    # Return all policy tables.
    all_tables = []
    try:
        # Get all the policies.
        policies = congress.policies_list(request)
    except Exception as e:
        LOG.error('Unable to get list of policies: %s' % e.message)
    else:
        try:
            for policy in policies:
                # Get all the tables in this policy.
                policy_name = policy['name']
                policy_tables = congress.policy_tables_list(request,
                                                            policy_name)
                # Get the names of the tables.
                datasource_tables = []
                for table in policy_tables:
                    table.set_id_as_name_if_empty()
                    table_name = table['name']
                    # Exclude service-derived tables.
                    if congress.TABLE_SEPARATOR not in table_name:
                        datasource_tables.append(table['name'])

                all_tables.append({'datasource': policy_name,
                                   'tables': datasource_tables})
        except Exception as e:
            LOG.error('Unable to get tables for policy "%s": %s' %
                      (policy_name, e.message))
    return all_tables


def _get_service_tables(request):
    # Return all service tables.
    all_tables = []
    try:
        # Get all the services.
        services = congress.datasources_list(request)
    except Exception as e:
        LOG.error('Unable to get list of data sources: %s' % e.message)
    else:
        try:
            for service in services:
                # Get all the tables in this service.
                service_id = service['id']
                service_tables = congress.datasource_tables_list(request,
                                                                 service_id)
                # Get the names of the tables.
                datasource_tables = []
                for table in service_tables:
                    table.set_id_as_name_if_empty()
                    datasource_tables.append(table['name'])

                all_tables.append({'datasource': service['name'],
                                   'tables': datasource_tables})
        except Exception as e:
            LOG.error('Unable to get tables for data source "%s": %s' %
                      (service_id, e.message))
    return all_tables


def get_datasource_tables(request):
    """Get names of all data source tables.

    Example:
    [
        {
            'datasource': 'classification',
            'tables': ['error']
        },
        {
            'datasource': 'neutronv2'
            'tables': ['networks', 'ports', ...]
        },
        ...
    ]
    """
    tables = _get_policy_tables(request)
    tables.extend(_get_service_tables(request))
    return tables


def get_datasource_columns(request):
    """Get of names of columns from all data sources.

    Example:
    [
        {
            'datasource': 'classification',
            'tables': [
                {
                    'table': 'error',
                    'columns': ['name']
                }
            ]
        },
        {
            'datasource': 'neutronv2',
            'tables': [
                {
                    'table': 'networks',
                    'columns':  ['id', 'tenant_id', ...],
                },
                ...
            ],
            ...
        },
        ...
    ]
    """
    all_columns = []

    # Get all the policy tables.
    policy_tables = _get_policy_tables(request)
    try:
        for policy in policy_tables:
            # Get all the columns in this policy. Unlike for the services,
            # there's currently no congress client API to get the schema for
            # all tables in a policy in a single call.
            policy_name = policy['datasource']
            tables = policy['tables']

            datasource_tables = []
            for table_name in tables:
                # Get all the columns in this policy table.
                schema = congress.policy_table_schema_get(request, policy_name,
                                                          table_name)
                columns = [c['name'] for c in schema['columns']]
                datasource_tables.append({'table': table_name,
                                          'columns': columns})

            all_columns.append({'datasource': policy_name,
                                'tables': datasource_tables})
    except Exception as e:
        LOG.error('Unable to get schema for policy "%s" table "%s": %s' %
                  (policy_name, table_name, e.message))

    try:
        # Get all the services.
        services = congress.datasources_list(request)
    except Exception as e:
        LOG.error('Unable to get list of data sources: %s' % e.message)
    else:
        try:
            for service in services:
                # Get the schema for this service.
                service_id = service['id']
                service_name = service['name']
                schema = congress.datasource_schema_get(request, service_id)

                datasource_tables = []
                for table in schema['tables']:
                    # Get the columns for this table.
                    columns = [c['name'] for c in table['columns']]
                    datasource_table = {'table': table['table_id'],
                                        'columns': columns}
                    datasource_tables.append(datasource_table)

                all_columns.append({'datasource': service_name,
                                    'tables': datasource_tables})
        except Exception as e:
            LOG.error('Unable to get schema for data source "%s": %s' %
                      (service_id, e.message))

    return all_columns



# Copyright (c) 2015 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from six.moves import range

from congress.datalog import base
from congress.datalog import compile
from congress.datalog import topdown
from congress.datalog import unify
from congress.datalog import utility
from congress import exception


##############################################################################
# Concrete Theory: Database
##############################################################################
class Database(topdown.TopDownTheory):
    class Proof(object):
        def __init__(self, binding, rule):
            self.binding = binding
            self.rule = rule

        def __str__(self):
            return "apply({}, {})".format(str(self.binding), str(self.rule))

        def __eq__(self, other):
            result = (self.binding == other.binding and
                      self.rule == other.rule)
            # LOG.debug("Pf: Comparing %s and %s: %s", self, other, result)
            # LOG.debug("Pf: %s == %s is %s",
            #     self.binding, other.binding, self.binding == other.binding)
            # LOG.debug("Pf: %s == %s is %s",
            #     self.rule, other.rule, self.rule == other.rule)
            return result

    class ProofCollection(object):
        def __init__(self, proofs):
            self.contents = list(proofs)

        def __str__(self):
            return '{' + ",".join(str(x) for x in self.contents) + '}'

        def __isub__(self, other):
            if other is None:
                return
            # LOG.debug("PC: Subtracting %s and %s", self, other)
            remaining = []
            for proof in self.contents:
                if proof not in other.contents:
                    remaining.append(proof)
            self.contents = remaining
            return self

        def __ior__(self, other):
            if other is None:
                return
            # LOG.debug("PC: Unioning %s and %s", self, other)
            for proof in other.contents:
                # LOG.debug("PC: Considering %s", proof)
                if proof not in self.contents:
                    self.contents.append(proof)
            return self

        def __getitem__(self, key):
            return self.contents[key]

        def __len__(self):
            return len(self.contents)

        def __ge__(self, iterable):
            for proof in iterable:
                if proof not in self.contents:
                    # LOG.debug("Proof %s makes %s not >= %s",
                    #     proof, self, iterstr(iterable))
                    return False
            return True

        def __le__(self, iterable):
            for proof in self.contents:
                if proof not in iterable:
                    # LOG.debug("Proof %s makes %s not <= %s",
                    #     proof, self, iterstr(iterable))
                    return False
            return True

        def __eq__(self, other):
            return self <= other and other <= self

    class DBTuple(object):
        def __init__(self, iterable, proofs=None):
            self.tuple = tuple(iterable)
            if proofs is None:
                proofs = []
            self.proofs = Database.ProofCollection(proofs)

        def __eq__(self, other):
            return self.tuple == other.tuple

        def __str__(self):
            return str(self.tuple) + str(self.proofs)

        def __len__(self):
            return len(self.tuple)

        def __getitem__(self, index):
            return self.tuple[index]

        def __setitem__(self, index, value):
            self.tuple[index] = value

        def match(self, atom, unifier):
            # LOG.debug("DBTuple matching %s against atom %s in %s",
            #     self, iterstr(atom.arguments), unifier)
            if len(self.tuple) != len(atom.arguments):
                return None
            changes = []
            for i in range(0, len(atom.arguments)):
                val, binding = unifier.apply_full(atom.arguments[i])
                # LOG.debug("val(%s)=%s at %s; comparing to object %s",
                #     atom.arguments[i], val, binding, self.tuple[i])
                if val.is_variable():
                    changes.append(binding.add(
                        val, compile.Term.create_from_python(self.tuple[i]),
                        None))
                else:
                    if val.name != self.tuple[i]:
                        unify.undo_all(changes)
                        return None
            return changes

    def __init__(self, name=None, abbr=None, theories=None, schema=None,
                 desc=None, owner=None):
        super(Database, self).__init__(
            name=name, abbr=abbr, theories=theories, schema=schema,
            desc=desc, owner=owner)
        self.data = {}
        self.kind = base.DATABASE_POLICY_TYPE

    def str2(self):
        def hash2str(h):
            s = "{"
            s += ", ".join(["{} : {}".format(str(key), str(h[key]))
                           for key in h])
            return s

        def hashlist2str(h):
            strings = []
            for key in h:
                s = "{} : ".format(key)
                s += '['
                s += ', '.join([str(val) for val in h[key]])
                s += ']'
                strings.append(s)
            return '{' + ", ".join(strings) + '}'

        return hashlist2str(self.data)

    def __eq__(self, other):
        return self.data == other.data

    def __sub__(self, other):
        def add_tuple(table, dbtuple):
            new = [table]
            new.extend(dbtuple.tuple)
            results.append(new)

        results = []
        for table in self.data:
            if table not in other.data:
                for dbtuple in self.data[table]:
                    add_tuple(table, dbtuple)
            else:
                for dbtuple in self.data[table]:
                    if dbtuple not in other.data[table]:
                        add_tuple(table, dbtuple)
        return results

    def __or__(self, other):
        def add_db(db):
            for table in db.data:
                for dbtuple in db.data[table]:
                    result.insert(compile.Literal.create_from_table_tuple(
                        table, dbtuple.tuple), proofs=dbtuple.proofs)
        result = Database()
        add_db(self)
        add_db(other)
        return result

    def __getitem__(self, key):
        # KEY must be a tablename
        return self.data[key]

    def content(self, tablenames=None):
        """Return a sequence of Literals representing all the table data."""
        results = []
        if tablenames is None:
            tablenames = self.data.keys()
        for table in tablenames:
            if table not in self.data:
                continue
            for dbtuple in self.data[table]:
                results.append(compile.Literal.create_from_table_tuple(
                    table, dbtuple.tuple))
        return results

    def is_noop(self, event):
        """Returns T if EVENT is a noop on the database."""
        # insert/delete same code but with flipped return values
        # Code below is written as insert, except noop initialization.
        if event.is_insert():
            noop = True
        else:
            noop = False
        if event.formula.table.table not in self.data:
            return not noop
        event_data = self.data[event.formula.table.table]
        raw_tuple = tuple(event.formula.argument_names())
        for dbtuple in event_data:
            if dbtuple.tuple == raw_tuple:
                if event.proofs <= dbtuple.proofs:
                    return noop
        return not noop

    def __contains__(self, formula):
        if not compile.is_atom(formula):
            return False
        if formula.table.table not in self.data:
            return False
        event_data = self.data[formula.table.table]
        raw_tuple = tuple(formula.argument_names())
        return any((dbtuple.tuple == raw_tuple for dbtuple in event_data))

    def explain(self, atom):
        if atom.table.table not in self.data or not atom.is_ground():
            return self.ProofCollection([])
        args = tuple([x.name for x in atom.arguments])
        for dbtuple in self.data[atom.table.table]:
            if dbtuple.tuple == args:
                return dbtuple.proofs

    def tablenames(self, body_only=False, include_builtin=False,
                   include_modal=True):
        """Return all table names occurring in this theory."""
        if body_only:
            return []
        return self.data.keys()

    # overloads for TopDownTheory so we can properly use the
    #    top_down_evaluation routines
    def defined_tablenames(self):
        return self.data.keys()

    def head_index(self, table, match_literal=None):
        if table not in self.data:
            return []
        return self.data[table]

    def head(self, thing):
        return thing

    def body(self, thing):
        return []

    def bi_unify(self, dbtuple, unifier1, atom, unifier2, theoryname):
        """THING1 is always a ground DBTuple and THING2 is always an ATOM."""
        return dbtuple.match(atom, unifier2)

    def atom_to_internal(self, atom, proofs=None):
        return atom.table.table, self.DBTuple(atom.argument_names(), proofs)

    def insert(self, atom, proofs=None):
        """Inserts ATOM into the DB.  Returns changes."""
        return self.modify(compile.Event(formula=atom, insert=True,
                                         proofs=proofs))

    def delete(self, atom, proofs=None):
        """Deletes ATOM from the DB.  Returns changes."""
        return self.modify(compile.Event(formula=atom, insert=False,
                                         proofs=proofs))

    def update(self, events):
        """Applies all of EVENTS to the DB.

        Each event is either an insert or a delete.
        """
        changes = []
        for event in events:
            changes.extend(self.modify(event))
        return changes

    def update_would_cause_errors(self, events):
        """Return a list of Policyxception.

        Return a list of PolicyException if we were
        to apply the events EVENTS to the current policy.
        """
        self.log(None, "update_would_cause_errors %s", utility.iterstr(events))
        errors = []
        for event in events:
            if not compile.is_atom(event.formula):
                errors.append(exception.PolicyException(
                    "Non-atomic formula is not permitted: {}".format(
                        str(event.formula))))
            else:
                errors.extend(compile.fact_errors(
                    event.formula, self.theories, self.name))
        return errors

    def modify(self, event):
        """Insert/Delete atom.

        Inserts/deletes ATOM and returns a list of changes that
        were caused. That list contains either 0 or 1 Event.
        """
        assert compile.is_atom(event.formula), "Modify requires Atom"
        atom = event.formula
        self.log(atom.table.table, "Modify: %s", atom)
        if self.is_noop(event):
            self.log(atom.table.table, "Event %s is a noop", event)
            return []
        if event.insert:
            self.insert_actual(atom, proofs=event.proofs)
        else:
            self.delete_actual(atom, proofs=event.proofs)
        return [event]

    def insert_actual(self, atom, proofs=None):
        """Workhorse for inserting ATOM into the DB.

        Along with proofs explaining how ATOM was computed from other tables.
        """
        assert compile.is_atom(atom), "Insert requires Atom"
        table, dbtuple = self.atom_to_internal(atom, proofs)
        self.log(table, "Insert: %s", atom)
        if table not in self.data:
            self.data[table] = [dbtuple]
            self.log(atom.table.table, "First tuple in table %s", table)
            return
        else:
            for existingtuple in self.data[table]:
                assert existingtuple.proofs is not None
                if existingtuple.tuple == dbtuple.tuple:
                    assert existingtuple.proofs is not None
                    existingtuple.proofs |= dbtuple.proofs
                    assert existingtuple.proofs is not None
                    return
            self.data[table].append(dbtuple)

    def delete_actual(self, atom, proofs=None):
        """Workhorse for deleting ATOM from the DB.

        Along with the proofs that are no longer true.
        """
        assert compile.is_atom(atom), "Delete requires Atom"
        self.log(atom.table.table, "Delete: %s", atom)
        table, dbtuple = self.atom_to_internal(atom, proofs)
        if table not in self.data:
            return
        for i in range(0, len(self.data[table])):
            existingtuple = self.data[table][i]
            if existingtuple.tuple == dbtuple.tuple:
                existingtuple.proofs -= dbtuple.proofs
                if len(existingtuple.proofs) == 0:
                    del self.data[table][i]
                return

    def policy(self):
        """Return the policy for this theory.

        No policy in this theory; only data.
        """
        return []

    def get_arity_self(self, tablename):
        if tablename not in self.data:
            return None
        if len(self.data[tablename]) == 0:
            return None
        return len(self.data[tablename][0].tuple)

    def content_string(self):
        s = ""
        for lit in self.content():
            s += str(lit) + '\n'
        return s + '\n'



# Copyright (c) 2016 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from oslo_log import log as logging
from oslo_serialization import jsonutils as json

LOG = logging.getLogger(__name__)


class DataServiceInfo(object):
    """Metadata for DataService on the DSE.

    Attributes:
        service_id: The ID of the service.
        node_id: The ID of the node the service is running on.
        published_tables: List of table IDs published by this service.
        subscribed_tables: List of table IDs this service subscribes to.
        rpc_endpoints_info: List of RPC endpoints exposed by this service.
    """
    MARSHALL_ATTRS = set(['service_id', 'node_id', 'published_tables',
                          'subscribed_tables', 'rpc_endpoints_info'])

    def __init__(self, service_id=None, node_id=None, published_tables=None,
                 subscribed_tables=None, rpc_endpoints_info=None):
        self.service_id = service_id
        self.node_id = node_id
        self.published_tables = published_tables or []
        self.subscribed_tables = subscribed_tables or []
        self.rpc_endpoints_info = rpc_endpoints_info or []

    def __str__(self):
        return self.__class__.__name__ + "<%s>" % self.service_id

    def __repr__(self):
        return self.__class__.__name__ + "(%s)" % dict.__repr__(vars(self))

    @classmethod
    def from_json(cls, json_string):
        return json.loads(json_string, object_hook=cls.from_dict)

    @classmethod
    def from_dict(cls, raw_dict):
        provided_keys = set(raw_dict.keys())
        if provided_keys != cls.MARSHALL_ATTRS:
            missing = cls.MARSHALL_ATTRS - provided_keys
            malformed = provided_keys - cls.MARSHALL_ATTRS
            msg = "Cannot construct %s from input:" % cls.__name__
            if missing:
                msg += " Missing keys: %s" % list(missing)
            if malformed:
                msg += " Malformed keys: %s" % list(malformed)
            raise KeyError(msg)
        ret = DataServiceInfo()
        for n in cls.MARSHALL_ATTRS:
            setattr(ret, n, raw_dict[n])
        return ret

    def to_dict(self):
        return dict([(k, getattr(self, k)) for k in self.MARSHALL_ATTRS])

    def to_json(self):
        return json.dumps(self.to_dict())


class DataService(object):
    """A unit of data and business logic that interfaces with the DSE.

    A DataService may publish tables, subscribe to tables, and/or expose
    RPCs on the DSE.  DataService instances are bound to a DseNode which is
    used for all inter-service communication.

    Attributes:
        service_id: A unique ID of the service.
        _published_tables_with_subscriber: A set of tables published by self
            that has subscribers
    """

    # TODO(pballand): make default methods for pub/subscribed tables
    def __init__(self, service_id):
        self.service_id = service_id
        self.node = None
        self._rpc_endpoints = [DataServiceEndPoints(self)]
        self._running = False
        self._published_tables_with_subscriber = set()

    def add_rpc_endpoint(self, endpt):
        self._rpc_endpoints.append(endpt)

    def rpc_endpoints(self):
        """Return list of RPC endpoint objects to be exposed for this service.

        A DataService may include zero or more RPC endpoints to be exposed
        by the DseNode.  Each endpoint object must be compatible with the
        oslo.messaging RPC Server.
        """
        return self._rpc_endpoints

    @property
    def status(self):
        return "OK"

    @property
    def info(self):
        # TODO(pballand): populate rpc_endpoints_info from rpc_endpoints
        return DataServiceInfo(
            service_id=self.service_id,
            node_id=self.node.node_id if self.node else None,
            published_tables=None,
            subscribed_tables=None,
            rpc_endpoints_info=None,
            )

    def start(self):
        """Start the DataService.

        This method is called by a DseNode before any RPCs are invoked.
        """
        assert self.node is not None
        self._running = True

    def stop(self):
        """Stop the DataService.

        This method is called by a DseNode when the DataService instance is
        no longer needed.  No RPCs will invoked on stopped DataServices.
        """
        assert self.node is not None
        self._running = False

    def wait(self):
        """Wait for processing to complete.

        After a call to stop(), the DataService may have some outstanding work
        that has not yet completed.  The wait() method blocks until all
        DataService processing is complete.
        """
        assert self.node is not None
        pass

    def rpc(self, service, action, kwargs=None):
        if kwargs is None:
            kwargs = {}
        return self.node.invoke_service_rpc(service, action, **kwargs)

    # Will be removed once the reference of node exists in api
    def get_datasources(self, filter_secret=False):
        return self.node.get_datasources(filter_secret)

    # Will be removed once the reference of node exists in api
    def get_datasource(self, datasource_id):
        return self.node.get_datasource(datasource_id)

    # Will be removed once the reference of node exists in api
    def add_datasource(self, **kwargs):
        return self.node.add_datasource(**kwargs)

    # Will be removed once the reference of node exists in api
    def delete_datasource(self, datasource):
        return self.node.delete_datasource(datasource)

    def publish(self, table, data):
        self.node.publish_table(self.service_id, table, data)

    def subscribe(self, service, table):
        data = self.node.subscribe_table(self.service_id, service, table)
        self.receive_data(service, table, data)

    def unsubscribe(self, service, table):
        self.node.unsubscribe_table(self.service_id, service, table)

    def receive_data(self, publisher, table, data):
        """Method called when publication data arrives.

           Instances will override this method.
        """
        data = self.node.to_set_of_tuples(data)
        self.last_msg = {}
        self.last_msg['data'] = data
        self.last_msg['publisher'] = publisher
        self.last_msg['table'] = table

    def subscription_list(self):
        """Method that returns subscription list.

        It returns list of tuple that represents the service's subscription.
        The tuple forms following format:
        (service_id, table_name).
        """
        result = []
        subscription = self.node.get_subscription(self.service_id)
        for target, tables in subscription.items():
            result.extend([(target, t) for t in tables])
        return result

    def subscriber_list(self):
        """Method that returns subscribers list.

        This feature is duplicated in the distributed architecture. So the
        method is defined only for backward compatibility.
        """
        LOG.info('subscriber_list is duplicated in the new architecture.')
        return []

    def get_snapshot(self, table):
        """Method that returns the current data for the given table.

           Should be overridden.
        """
        raise NotImplementedError(
            "get_snapshot is not implemented in the '%s' class." %
            self.service_id)


class DataServiceEndPoints (object):
    def __init__(self, service):
        self.service = service

    def get_snapshot(self, context, table):
        """Function called on a node when an RPC request is sent."""
        try:
            return self.service.get_snapshot(table)
        except AttributeError:
            pass



# Copyright 2014 Plexxi, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import eventlet
from eventlet import greenthread
from eventlet import hubs
eventlet.monkey_patch()
from oslo_log import log as logging
from oslo_utils import strutils

from congress.dse import d6message
from congress.dse import dataobj

LOG = logging.getLogger(__name__)


class deepSix(greenthread.GreenThread):
    def __init__(self, name, keys, inbox=None, dataPath=None):
        hub = hubs.get_hub()
        greenthread.GreenThread.__init__(self, hub.greenlet)
        g = self

        self.name = name
        keyList = []

        for k in keys:
            keyList.append(k)
            localk = "local." + k
            keyList.append(localk)

        keyList.append("allservices")
        keyList.append("local.allservices")

        self.keys = keyList

        self.running = True

        self.pubdata = {}
        self.subdata = {}
        self.subscriberCorrelationUuids = set()
        self.scheduuids = set()
        self.timerThreads = []

        # Necessary for deepSix objects that don't get initialized with an
        # inbox
        self.inbox = None

        if inbox:
            self.inbox = inbox
            self.dataPath = dataPath

        hub.schedule_call_global(0, g.switch, g._loop, [], {})

        keyargs = {}
        keyargs['keys'] = self.keys

        self.publish("routeKeys", keyargs)

    def send(self, msg):
        # TODO(thinrichs): reduce how often sub messages
        #   get sent so we can re-enable this
        # if msg.type == 'sub':
        #     self.log_info("sending SUB msg %s", msg)
        # else:
        #     self.log_debug("sending msg %s", msg)
        self.dataPath.put_nowait(msg)

    def schedule(self, msg, scheduuid, interval):
        if scheduuid in self.scheduuids:

            if msg.type == 'pub':
                msg.updatebody(self.pubdata[msg.dataindex].get())

            self.send(msg)

            ev = eventlet.spawn_after(interval,
                                      self.schedule,
                                      msg,
                                      scheduuid,
                                      interval)
            self.timerThreads.append(ev)
        else:
            self.log_debug("stop scheduling a message: %s", msg)

    def getSubData(self, corrId, sender=""):
        if corrId in self.subdata:
            if sender:
                return self.subdata[corrId].getData(sender)
            else:
                return self.subdata[corrId].getAllData()

    def reqtimeout(self, corrId):
        if corrId in self.subdata:
            del self.subdata[corrId]

    def inreq(self, msg):
        corruuid = msg.correlationId
        dataindex = msg.header['dataindex']

        if dataindex == "pubdata":
            newmsg = d6message.d6msg(key=msg.replyTo,
                                     replyTo=self.name,
                                     correlationId=msg.correlationId,
                                     type="rep",
                                     dataindex=dataindex,
                                     body=dataobj.dataObject(self.pubdata))
            self.send(newmsg)

        elif dataindex == "subdata":
            newmsg = d6message.d6msg(key=msg.replyTo,
                                     replyTo=self.name,
                                     correlationId=msg.correlationId,
                                     type="rep",
                                     dataindex=dataindex,
                                     body=dataobj.dataObject(self.subdata))
            self.send(newmsg)

        elif hasattr(self, 'reqhandler'):
            self.pubdata[dataindex] = dataobj.pubData(dataindex, msg.body)
            self.pubdata[dataindex].requesters[msg.replyTo] = corruuid
            self.reqhandler(msg)

        else:
            self.log_exception("Received a request but have no handler: %s",
                               msg)

    def inpull(self, msg):
        # self.log_debug("received PULL msg: %s", msg)
        dataindex = msg.header['dataindex']

        if dataindex in self.pubdata:

            reply = d6message.d6msg(replyTo=self.name,
                                    type="rep",
                                    body=self.pubdata[dataindex].get(),
                                    srcmsg=msg)
            self.send(reply)

        else:
            self.pubdata[dataindex] = dataobj.pubData(dataindex, msg.body)
            self.subhandler(msg)

        self.pubdata[dataindex].addsubscriber(
            msg.replyTo, "pull", msg.correlationId)

    def incmd(self, msg):
        # self.log_debug("received CMD msg: %s", msg)
        corruuid = msg.correlationId
        dataindex = msg.header['dataindex']

        if corruuid not in self.pubdata:
            self.pubdata[corruuid] = dataobj.pubData(dataindex, msg.body)
            self.pubdata[corruuid].requesters[msg.replyTo] = corruuid
            self.cmdhandler(msg)

    def insub(self, msg):
        # self.log_info("received SUB msg: %s", msg)
        corruuid = msg.correlationId
        dataindex = msg.header['dataindex']
        sender = msg.replyTo

        if corruuid not in self.subscriberCorrelationUuids:

            if dataindex not in self.pubdata:
                self.pubdata[dataindex] = dataobj.pubData(dataindex, msg.body)
            # always call subhandler so subclass has a chance to know more
            # about the subscription
            if hasattr(self, "subhandler"):
                self.subhandler(msg)

            self.pubdata[dataindex].addsubscriber(sender, "push", corruuid)
            self.subscriberCorrelationUuids.add(corruuid)
            self.push(dataindex, sender, type='sub')

    def inunsub(self, msg):
        # self.log_info("received UNSUB msg: %s", msg)
        dataindex = msg.header['dataindex']

        if hasattr(self, 'unsubhandler'):
            if self.unsubhandler(msg):
                if dataindex in self.pubdata:
                    self.pubdata[dataindex].removesubscriber(msg.replyTo)
        else:
            if dataindex in self.pubdata:
                self.pubdata[dataindex].removesubscriber(msg.replyTo)

        # release resource if no more subscribers for this dataindex
        if self.pubdata[dataindex].getsubscribers() == []:
            self.pubdata.discard(dataindex)

    def inshut(self, msg):
        """Shut down this data service."""
        # self.log_warning("received SHUT msg: %s", msg)

        for corruuid in self.subdata:
            self.unsubscribe(corrId=corruuid)

        for ev in self.timerThreads:
            try:
                ev.kill()
            except Exception as errmsg:
                self.log("error stopping timer thread: %s", errmsg)

        self.running = False

        self.keys = {}
        keydata = {}
        keydata['keys'] = {}
        self.publish("routeKeys", keydata)

    def inpubrep(self, msg):
        # self.log_debug("received PUBREP msg: %s", msg)
        corruuid = msg.correlationId
        sender = msg.replyTo

        if corruuid in self.scheduuids:
            self.scheduuids.remove(corruuid)

        if corruuid in self.subdata:
            callback = self.subdata[corruuid].callback

            if msg.type in ['pub', 'rep']:
                if callback:
                    scrubbed = callback(msg)
                    if scrubbed:
                        self.subdata[corruuid].update(
                            sender, dataobj.dataObject(scrubbed))
        else:
            self.unsubscribe(corrId=corruuid)

    def request(
            self,
            key,
            dataindex,
            corrId="",
            callback=None,
            interval=0,
            timer=30,
            args={}):
        msg = d6message.d6msg(key=key,
                              replyTo=self.name,
                              correlationId=corrId,
                              type="req",
                              dataindex=dataindex,
                              body=args)

        corruuid = msg.correlationId
        self.subdata[corruuid] = dataobj.subData(key, dataindex,
                                                 corruuid, callback)

        if interval:
            self.scheduuids.add(corruuid)
            self.schedule(msg, corruuid, interval)
        else:

            self.send(msg)

            if timer:
                self.timerThreads.append(
                    eventlet.spawn_after(timer,
                                         self.reqtimeout,
                                         corruuid))

    def reply(self, dataindex, newdata="", delete=True):
        for requester in self.pubdata[dataindex].requesters:

            msg = d6message.d6msg(key=requester,
                                  replyTo=self.name,
                                  correlationId=self.pubdata[dataindex]
                                  .requesters[requester],
                                  type="rep",
                                  dataindex=self.pubdata[dataindex].dataindex)

            if newdata:
                msg.body = dataobj.dataObject(newdata)
            else:
                msg.body = self.pubdata[dataindex].get()
            # self.log_debug("REPLY body: %s", msg.body)

            self.send(msg)

        if delete:

            del self.pubdata[dataindex]

    def prepush_processor(self, data, dataindex, type=None):
        """Pre-processing the data before publish.

        Given the DATA to be published, returns the data actually put
        on the wire.  Can be overloaded.
        """
        return data

    def reserved_dataindex(self, dataindex):
        """Returns True if DATAINDEX is one of those reserved by deepsix."""
        return dataindex in ('routeKeys', 'pubdata', 'subdata')

    def push(self, dataindex, key="", type=None):
        """Send data for DATAINDEX and KEY to subscribers/requesters."""
        self.log_debug("pushing dataindex %s to subscribers %s "
                       "and requesters %s ", dataindex,
                       self.pubdata[dataindex].subscribers,
                       self.pubdata[dataindex].requesters)

        # bail out if there are no requesters/subscribers
        if (len(self.pubdata[dataindex].requesters) == 0 and
                len(self.pubdata[dataindex].subscribers) == 0):
            self.log_debug("no requesters/subscribers; not sending")
            return

        # give prepush hook chance to morph data
        if self.reserved_dataindex(dataindex):
            data = self.pubdata[dataindex].get()
            # bail out if no data to send
            if data is None:
                return
        else:
            # .get() returns dataObject
            data = self.prepush_processor(self.pubdata[dataindex].get().data,
                                          dataindex,
                                          type=type)
            # bail out if prepush hook said there's no data
            if data is None:
                return
            data = dataobj.dataObject(data)

        # send to subscribers/requestors
        if self.pubdata[dataindex].subscribers:

            if key:
                msg = d6message.d6msg(key=key,
                                      replyTo=self.name,
                                      correlationId=self.pubdata[dataindex]
                                      .subscribers[key]['correlationId'],
                                      type="pub",
                                      dataindex=dataindex,
                                      body=data)
                self.send(msg)
            else:
                subscribers = self.pubdata[dataindex].getsubscribers()
                for subscriber in subscribers:

                    if subscribers[subscriber]['type'] == "push":
                        corId = subscribers[subscriber]['correlationId']
                        msg = d6message.d6msg(key=subscriber,
                                              replyTo=self.name,
                                              correlationId=corId,
                                              type="pub",
                                              dataindex=dataindex,
                                              body=data)

                        self.send(msg)

        if self.pubdata[dataindex].requesters:
            if key:
                msg = d6message.d6msg(key=key,
                                      replyTo=self.name,
                                      correlationId=self.pubdata[dataindex].
                                      requesters[key],
                                      type="rep",
                                      dataindex=dataindex,
                                      body=self.pubdata[dataindex].get())
                self.send(msg)
                del self.pubdata[dataindex].requesters[key]
            else:
                for requester in self.pubdata[dataindex].requesters.keys():
                    corId = self.pubdata[dataindex].requesters[requester]
                    msg = d6message.d6msg(key=requester,
                                          replyTo=self.name,
                                          correlationId=corId,
                                          type="rep",
                                          dataindex=dataindex,
                                          body=self.pubdata[dataindex].get())
                    self.send(msg)
                    del self.pubdata[dataindex].requesters[requester]

    def subscribe(
            self,
            key,
            dataindex,
            corrId="",
            callback=None,
            pull=False,
            interval=30,
            args={}):
        """Subscribe to a DATAINDEX for a given KEY."""
        self.log_debug("subscribed to %s with dataindex %s", key, dataindex)

        msg = d6message.d6msg(key=key,
                              replyTo=self.name,
                              correlationId=corrId,
                              dataindex=dataindex,
                              body=args)
        if pull:
            msg.type = 'pull'
        else:
            msg.type = 'sub'

        corruuid = msg.correlationId

        self.subdata[corruuid] = dataobj.subData(key, dataindex,
                                                 corruuid, callback)
        self.scheduuids.add(corruuid)
        self.schedule(msg, corruuid, interval)

        return corruuid

    def unsubscribe(self, key="", dataindex="", corrId=""):
        """Unsubscribe self from DATAINDEX for KEY."""
        self.log_debug("unsubscribed to %s with dataindex %s", key, dataindex)
        if corrId:
            if corrId in self.scheduuids:
                self.scheduuids.remove(corrId)
            if corrId in self.subdata:
                key = self.subdata[corrId].key
                dataindex = self.subdata[corrId].dataindex
                del self.subdata[corrId]

            msg = d6message.d6msg(key=key,
                                  replyTo=self.name,
                                  correlationId=corrId,
                                  type='unsub',
                                  dataindex=dataindex)

            self.send(msg)

        elif key and dataindex:

            for corruuid in self.subdata.copy().keys():
                # copy to avoid undefined behavior w changing dict during iter

                if (key == self.subdata[corruuid].key and
                        dataindex == self.subdata[corruuid].dataindex):

                    if corruuid in self.scheduuids:
                        self.scheduuids.remove(corruuid)

                    del self.subdata[corruuid]

                    msg = d6message.d6msg(key=key,
                                          replyTo=self.name,
                                          correlationId=corruuid,
                                          type='unsub',
                                          dataindex=dataindex)
                    self.send(msg)

        return

    def command(
            self,
            key,
            command,
            corrId="",
            callback=None,
            timer=30,
            args={}):
        msg = d6message.d6msg(key=key,
                              replyTo=self.name,
                              type="cmd",
                              correlationId=corrId,
                              dataindex=command,
                              body=args)

        corruuid = msg.correlationId

        self.subdata[corruuid] = dataobj.subData(key, command,
                                                 corruuid, callback)

        self.send(msg)

        if timer:
            self.timerThreads.append(
                eventlet.spawn_after(timer,
                                     self.reqtimeout,
                                     corruuid))

    def publish(self, dataindex, newdata, key=''):
        self.log_debug("publishing to dataindex %s with data %s",
                       dataindex, strutils.mask_password(newdata, "****"))
        if dataindex not in self.pubdata:
            self.pubdata[dataindex] = dataobj.pubData(dataindex)

        self.pubdata[dataindex].update(newdata)

        self.push(dataindex, type='pub')

    def receive(self, msg):
        if msg.type == 'sub':
            self.insub(msg)
        elif msg.type == 'unsub':
            self.inunsub(msg)
        elif msg.type == 'pub':
            self.inpubrep(msg)
        elif msg.type == 'req':
            self.inreq(msg)
        elif msg.type == 'rep':
            self.inpubrep(msg)
        elif msg.type == 'pull':
            self.inpull(msg)
        elif msg.type == 'shut':
            self.inshut(msg)
        elif msg.type == 'cmd':
            if hasattr(self, 'cmdhandler'):
                self.incmd(msg)
        else:
            assert False, "{} received message of unknown type {}: {}".format(
                self.name, msg.type, str(msg))

    def _loop(self):

        # self.running will be set to False when processing a shutdown a
        # message
        while self.running:
            if self.inbox:
                msg = self.inbox.get()
                self.receive(msg)
                self.inbox.task_done()
            else:
                # in test cases some deepSix instances are initialized
                # without an inbox, this prevents a busy wait state
                eventlet.sleep(1)

    def subscription_list(self):
        """Return a list version of subscriptions."""
        return [(x.key, x.dataindex) for x in self.subdata.values()]

    def subscriber_list(self):
        """Return a list version of subscribers."""
        result = []
        for pubdata in self.pubdata.values():
            for subscriber in pubdata.subscribers:
                result.append((subscriber, pubdata.dataindex))

    def log(self, msg, *args):
        self.log_debug(msg, *args)

    def log_debug(self, msg, *args):
        msg = "%s:: %s" % (self.name, msg)
        LOG.debug(msg, *args)

    def log_info(self, msg, *args):
        msg = "%s:: %s" % (self.name, msg)
        LOG.info(msg, *args)

    def log_warning(self, msg, *args):
        msg = "%s:: %s" % (self.name, msg)
        LOG.warning(msg, *args)

    def log_error(self, msg, *args):
        msg = "%s:: %s" % (self.name, msg)
        LOG.error(msg, *args)

    def log_exception(self, msg, *args):
        msg = "%s:: %s" % (self.name, msg)
        LOG.exception(msg, *args)



# Copyright (c) 2014 Marist SDN Innovation lab Joint with Plexxi Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import json

try:
    from plexxi.core.api.binding import AffinityGroup
    from plexxi.core.api.binding import Job
    from plexxi.core.api.binding import PhysicalPort
    from plexxi.core.api.binding import PlexxiSwitch
    from plexxi.core.api.binding import VirtualizationHost
    from plexxi.core.api.binding import VirtualMachine
    from plexxi.core.api.binding import VirtualSwitch
    from plexxi.core.api.binding import VmwareVirtualMachine
    from plexxi.core.api.session import CoreSession
except ImportError:
    pass

from oslo_config import cfg
from oslo_log import log as logging
import requests

from congress.datasources import constants
from congress.datasources import datasource_driver
from congress.managers import datasource as datasource_mgr

LOG = logging.getLogger(__name__)


def d6service(name, keys, inbox, datapath, args):
    """This method is called by d6cage to create a dataservice instance."""

    return PlexxiDriver(name, keys, inbox, datapath, args)


class PlexxiDriver(datasource_driver.PollingDataSourceDriver,
                   datasource_driver.ExecutionDriver):
    HOSTS = "hosts"
    HOST_MACS = HOSTS + '.macs'
    HOST_GUESTS = HOSTS + '.guests'
    VMS = "vms"
    VM_MACS = VMS + '.macs'
    AFFINITIES = "affinities"
    VSWITCHES = "vswitches"
    VSWITCHES_MACS = VSWITCHES + '.macs'
    VSWITCHES_HOSTS = VSWITCHES + '.hosts'
    PLEXXISWITCHES = "plexxiswitches"
    PLEXXISWITCHES_MACS = PLEXXISWITCHES + '.macs'
    PORTS = "ports"
    NETWORKLINKS = "networklinks"

    def __init__(self, name='', keys='', inbox=None, datapath=None, args=None,
                 session=None):
        super(PlexxiDriver, self).__init__(name, keys, inbox, datapath, args)
        datasource_driver.ExecutionDriver.__init__(self)
        self.exchange = session
        self.creds = args
        self.raw_state = {}
        try:
            self.unique_names = self.string_to_bool(args['unique_names'])
        except KeyError:
            LOG.warning("unique_names has not been configured, "
                        "defaulting to False.")
            self.unique_names = False
        port = str(cfg.CONF.bind_port)
        host = str(cfg.CONF.bind_host)
        self.headers = {'content-type': 'application/json'}
        self.name_cooldown = False
        self.api_address = "http://" + host + ":" + port + "/v1"
        self.name_rule_needed = True
        if str(cfg.CONF.auth_strategy) == 'keystone':
            if 'keystone_pass' not in args:
                LOG.error("Keystone is enabled, but a password was not " +
                          "provided. All automated API calls are disabled")
                self.unique_names = False
                self.name_rule_needed = False
            elif 'keystone_user' not in args:
                LOG.error("Keystone is enabled, but a username was not " +
                          "provided. All automated API calls are disabled")
                self.unique_names = False
                self.name_rule_needed = False
            else:
                self.keystone_url = str(cfg.CONF.keystone_authtoken.auth_uri)
                self.keystoneauth()
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'plexxi'
        result['description'] = ('Datasource driver that interfaces with '
                                 'PlexxiCore.')
        result['config'] = {'auth_url': constants.REQUIRED,  # PlexxiCore url
                            'username': constants.REQUIRED,
                            'password': constants.REQUIRED,
                            'poll_time': constants.OPTIONAL,
                            'tenant_name': constants.REQUIRED,
                            'unique_names': constants.OPTIONAL,
                            'keystone_pass': constants.OPTIONAL,
                            'keystone_user': constants.OPTIONAL}
        result['secret'] = ['password']
        return result

    def update_from_datasource(self):
        """Called when it is time to pull new data from this datasource.

        Pulls lists of objects from PlexxiCore, if the data does not match
        the correspondig table in the driver's raw state or has not yet been
        added to the state, the driver calls methods to parse this data.

        Once all data has been updated,sets
        self.state[tablename] = <list of tuples of strings/numbers>
        for every tablename exported by PlexxiCore.
        """

        # Initialize instance variables that get set during update
        self.hosts = []
        self.mac_list = []
        self.guest_list = []
        self.plexxi_switches = []
        self.affinities = []
        self.vswitches = []
        self.vms = []
        self.vm_macs = []
        self.ports = []
        self.network_links = []

        if self.exchange is None:
            self.connect_to_plexxi()

        # Get host data from PlexxiCore
        hosts = VirtualizationHost.getAll(session=self.exchange)
        if (self.HOSTS not in self.state or
           hosts != self.raw_state[self.HOSTS]):
            self._translate_hosts(hosts)
            self.raw_state[self.HOSTS] = hosts
        else:
            self.hosts = self.state[self.HOSTS]
            self.mac_list = self.state[self.HOST_MACS]
            self.guest_list = self.state[self.HOST_GUESTS]

        # Get PlexxiSwitch Data from PlexxiCore
        plexxiswitches = PlexxiSwitch.getAll(session=self.exchange)
        if (self.PLEXXISWITCHES not in self.state or
           plexxiswitches != self.raw_state[self.PLEXXISWITCHES]):
            self._translate_pswitches(plexxiswitches)
            self.raw_state[self.PLEXXISWITCHES] = plexxiswitches
        else:
            self.plexxi_switches = self.state[self.PLEXXISWITCHES]

        # Get affinity data from PlexxiCore
        affinities = AffinityGroup.getAll(session=self.exchange)
        if (self.AFFINITIES not in self.state or
           affinities != self.raw_state[self.AFFINITIES]):
            if AffinityGroup.getCount(session=self.exchange) == 0:
                self.state[self.AFFINITIES] = ['No Affinities found']
            else:
                self._translate_affinites(affinities)
            self.raw_state[self.AFFINITIES] = affinities
        else:
            self.affinties = self.state[self.AFFINITIES]

        # Get vswitch data from PlexxiCore
        vswitches = VirtualSwitch.getAll(session=self.exchange)
        if (self.VSWITCHES not in self.state or
           vswitches != self.raw_state[self.VSWITCHES]):
            self._translate_vswitches(vswitches)
            self.raw_state[self.VSWITCHES] = vswitches
        else:
            self.vswitches = self.state[self.VSWITCHES]

        # Get virtual machine data from PlexxiCore
        vms = VirtualMachine.getAll(session=self.exchange)
        if (self.VMS not in self.state or
           vms != self.raw_state[self.VMS]):
            self._translate_vms(vms)
            self.raw_state[self.VMS] = set(vms)
        else:
            self.vms = self.state[self.VMS]
            self.vm_macs = self.state[self.VMS_MACS]
        # Get port data from PlexxiCore
        ports = PhysicalPort.getAll(session=self.exchange)
        if(self.PORTS not in self.state or
           ports != self.raw_state[self.PORTS]):
            self._translate_ports(ports)
            self.raw_state[self.PORTS] = set(ports)
        else:
            self.ports = self.state[self.PORTS]
            self.network_links = self.state[self.NETWORKLINKS]

        LOG.debug("Setting Plexxi State")
        self.state = {}
        self.state[self.HOSTS] = set(self.hosts)
        self.state[self.HOST_MACS] = set(self.mac_list)
        self.state[self.HOST_GUESTS] = set(self.guest_list)
        self.state[self.PLEXXISWITCHES] = set(self.plexxi_switches)
        self.state[self.PLEXXISWITCHES_MACS] = set(self.ps_macs)
        self.state[self.AFFINITIES] = set(self.affinities)
        self.state[self.VSWITCHES] = set(self.vswitches)
        self.state[self.VSWITCHES_MACS] = set(self.vswitch_macs)
        self.state[self.VSWITCHES_HOSTS] = set(self.vswitch_hosts)
        self.state[self.VMS] = set(self.vms)
        self.state[self.VM_MACS] = set(self.vm_macs)
        self.state[self.PORTS] = set(self.ports)
        self.state[self.NETWORKLINKS] = set(self.network_links)

        # Create Rules
        if self.name_rule_needed is True:
            if self.name_rule_check() is True:
                self.name_rule_create()
            else:
                self.name_rule_needed = False
        # Act on Policy
        if self.unique_names is True:
            if not self.name_cooldown:
                self.name_response()
            else:
                self.name_cooldown = False

    @classmethod
    def get_schema(cls):
        """Creates a table schema for incoming data from PlexxiCore.

        Returns a dictionary map of tablenames corresponding to column names
        for that table. Both tableNames and columnnames are strings.
        """

        d = {}
        d[cls.HOSTS] = ("uuid", "name", "mac_count", "vmcount")
        d[cls.HOST_MACS] = ("Host_uuid", "Mac_Address")
        d[cls.HOST_GUESTS] = ("Host_uuid", "VM_uuid")
        d[cls.VMS] = ("uuid", "name", "host_uuid", "ip", "mac_count")
        d[cls.VM_MACS] = ("vmID", "Mac_Address")
        d[cls.AFFINITIES] = ("uuid", "name")
        d[cls.VSWITCHES] = ("uuid", "host_count", "vnic_count")
        d[cls.VSWITCHES_MACS] = ("vswitch_uuid", "Mac_Address")
        d[cls.VSWITCHES_HOSTS] = ("vswitch_uuid", "hostuuid")
        d[cls.PLEXXISWITCHES] = ("uuid", "ip", "status")
        d[cls.PLEXXISWITCHES_MACS] = ("Switch_uuid", "Mac_Address")
        d[cls.PORTS] = ("uuid", "name")
        d[cls.NETWORKLINKS] = ("uuid", "name", "port_uuid", "start_uuid",
                               "start_name", "stop_uuid", "stop_name")
        return d

    def _translate_hosts(self, hosts):
        """Translates data about Hosts from PlexxiCore for Congress.

        Responsible for the states 'hosts','hosts.macs' and 'hosts.guests'
        """

        row_keys = self.get_column_map(self.HOSTS)
        hostlist = []
        maclist = []
        vm_uuids = []
        for host in hosts:
            row = ['None'] * (max(row_keys.values()) + 1)
            hostID = host.getForeignUuid()
            row[row_keys['uuid']] = hostID
            row[row_keys['name']] = host.getName()
            pnics = host.getPhysicalNetworkInterfaces()
            if pnics:
                for pnic in pnics:
                    mac = str(pnic.getMacAddress())
                    tuple_mac = (hostID, mac)
                    maclist.append(tuple_mac)
            mac_count = len(maclist)
            if (mac_count > 0):
                row[row_keys['mac_count']] = mac_count
            vmCount = host.getVirtualMachineCount()
            row[row_keys['vmcount']] = vmCount
            if vmCount != 0:
                vms = host.getVirtualMachines()
                for vm in vms:
                    tuple_vmid = (hostID, vm.getForeignUuid())
                    vm_uuids.append(tuple_vmid)
            hostlist.append(tuple(row))
        self.hosts = hostlist
        self.mac_list = maclist
        self.guest_list = vm_uuids

    def _translate_pswitches(self, plexxi_switches):
        """Translates data on Plexxi Switches from PlexxiCore for Congress.

        Responsible for state 'Plexxi_swtiches' and 'Plexxi_switches.macs'
        """

        row_keys = self.get_column_map(self.PLEXXISWITCHES)
        pslist = []
        maclist = []
        for switch in plexxi_switches:
            row = ['None'] * (max(row_keys.values()) + 1)
            psuuid = str(switch.getUuid())
            row[row_keys['uuid']] = psuuid
            psip = str(switch.getIpAddress())
            row[row_keys['ip']] = psip
            psstatus = str(switch.getStatus())
            row[row_keys['status']] = psstatus
            pnics = switch.getPhysicalNetworkInterfaces()
            for pnic in pnics:
                mac = str(pnic.getMacAddress())
                macrow = [psuuid, mac]
                maclist.append(tuple(macrow))
            pslist.append(tuple(row))
        self.plexxi_switches = pslist
        self.ps_macs = maclist

    def _translate_affinites(self, affinites):
        """Translates data about affinites from PlexxiCore for Congress.

        Responsible for state 'affinities'
        """

        row_keys = self.get_column_map(self.AFFINITIES)
        affinitylist = []
        for affinity in affinites:
            row = ['None'] * (max(row_keys.values()) + 1)
            uuid = str(affinity.getUuid())
            row[row_keys['uuid']] = uuid
            row[row_keys['name']] = affinity.getName()
            affinitylist.append(tuple(row))
        self.affinities = affinitylist

    def _translate_vswitches(self, vswitches):
        """Translates data about vswitches from PlexxiCore for Congress.

        Responsible for states vswitchlist,vswitch_macs,vswitch_hosts
        """

        # untested
        row_keys = self.get_column_map(self.VSWITCHES)
        vswitchlist = []
        tuple_macs = []
        vswitch_host_list = []
        for vswitch in vswitches:
            row = ['None'] * (max(row_keys.values()) + 1)
            vswitchID = vswitch.getForeignUuid()
            row[row_keys['uuid']] = vswitchID
            vSwitchHosts = vswitch.getVirtualizationHosts()
            try:
                host_count = len(vSwitchHosts)
            except TypeError:
                host_count = 0
            row[row_keys['host_count']] = host_count
            if host_count != 0:
                for host in vSwitchHosts:
                    hostuuid = host.getForeignUuid()
                    hostrow = [vswitchID, hostuuid]
                    vswitch_host_list.append(tuple(hostrow))
            vswitch_vnics = vswitch.getVirtualNetworkInterfaces()
            try:
                vnic_count = len(vswitch_vnics)
            except TypeError:
                vnic_count = 0
            row[row_keys['vnic_count']] = vnic_count
            if vnic_count != 0:
                for vnic in vswitch_vnics:
                    mac = vnic.getMacAddress()
                    macrow = [vswitchID, str(mac)]
                    tuple_macs.append(tuple(macrow))
            vswitchlist.append(tuple(row))
        self.vswitches = vswitchlist
        self.vswitch_macs = tuple_macs
        self.vswitch_hosts = vswitch_host_list

    def _translate_vms(self, vms):
        """Translate data on VMs from PlexxiCore for Congress.

        Responsible for states 'vms' and 'vms.macs'
        """

        row_keys = self.get_column_map(self.VMS)
        vmlist = []
        maclist = []
        for vm in vms:
            row = ['None'] * (max(row_keys.values()) + 1)
            vmID = vm.getForeignUuid()
            row[row_keys['uuid']] = vmID
            vmName = vm.getName()
            row[row_keys['name']] = vmName
            try:
                vmhost = vm.getVirtualizationHost()
                vmhostuuid = vmhost.getForeignUuid()
                row[row_keys['host_uuid']] = vmhostuuid
            except AttributeError:
                LOG.debug("The host for " + vmName + " could not be found")
            vmIP = vm.getIpAddress()
            if vmIP:
                row[row_keys['ip']] = vmIP
            vmVnics = vm.getVirtualNetworkInterfaces()
            mac_count = 0
            for vnic in vmVnics:
                mac = str(vnic.getMacAddress())
                tuple_mac = (vmID, mac)
                maclist.append(tuple_mac)
                mac_count += 1
            row[row_keys['mac_count']] = mac_count
            vmlist.append(tuple(row))
        self.vms = vmlist
        self.vm_macs = maclist

    def _translate_ports(self, ports):
        """Translate data about ports from PlexxiCore for Congress.

        Responsible for states 'ports' and 'ports.links'
        """

        row_keys = self.get_column_map(self.PORTS)
        link_keys = self.get_column_map(self.NETWORKLINKS)
        port_list = []
        link_list = []
        for port in ports:
            row = ['None'] * (max(row_keys.values()) + 1)
            portID = str(port.getUuid())
            row[row_keys['uuid']] = portID
            portName = str(port.getName())
            row[row_keys['name']] = portName
            links = port.getNetworkLinks()
            if links:
                link_keys = self.get_column_map(self.NETWORKLINKS)
                for link in links:
                    link_row = self._translate_network_link(link, link_keys,
                                                            portID)
                    link_list.append(tuple(link_row))
            port_list.append(tuple(row))
        self.ports = port_list
        self.network_links = link_list

    def _translate_network_link(self, link, row_keys, sourcePortUuid):
        """Translates data about network links from PlexxiCore for Congress.

        Subfunction of translate_ports,each  handles a set of network links
        attached to a port. Directly responsible for the state of
        'ports.links'
        """

        row = ['None'] * (max(row_keys.values()) + 1)
        linkID = str(link.getUuid())
        row[row_keys['uuid']] = linkID
        row[row_keys['port_uuid']] = sourcePortUuid
        linkName = str(link.getName())
        row[row_keys['name']] = linkName
        linkStartObj = link.getStartNetworkInterface()
        linkStartName = str(linkStartObj.getName())
        row[row_keys['start_name']] = linkStartName
        linkStartUuid = str(linkStartObj.getUuid())
        row[row_keys['start_uuid']] = linkStartUuid
        linkStopObj = link.getStopNetworkInterface()
        linkStopUuid = str(linkStopObj.getUuid())
        row[row_keys['stop_uuid']] = linkStopUuid
        linkStopName = str(linkStopObj.getName())
        row[row_keys['stop_name']] = linkStopName
        return row

    def string_to_bool(self, string):
        """Used for parsing boolean variables stated in datasources.conf."""

        string = string.strip()
        s = string.lower()
        if s in['true', 'yes', 'on']:
            return True
        else:
            return False

    def connect_to_plexxi(self):
        """Connect to PlexxiCore.

        Create a CoreSession connecting congress to PlexxiCore using
        credentials provided in datasources.conf
        """

        if 'auth_url' not in self.creds:
            LOG.error("Plexxi url not supplied. Could not start Plexxi" +
                      "connection driver")
        if 'username' not in self.creds:
            LOG.error("Plexxi username not supplied. Could not start " +
                      "Plexxi connection driver")
        if 'password' not in self.creds:
            LOG.error("Plexxi password not supplied. Could not start " +
                      "Plexxi connection driver")
        try:
            self.exchange = CoreSession.connect(
                baseUrl=self.creds['auth_url'],
                allowUntrusted=True,
                username=self.creds['username'],
                password=self.creds['password'])
        except requests.exceptions.HTTPError as error:
            if (int(error.response.status_code) == 401 or
                    int(error.response.status_code) == 403):
                msg = ("Incorrect username/password combination. Passed" +
                       "in username was " + self.creds['username'])

                raise Exception(requests.exceptions.HTTPErrror(msg))
            else:
                raise Exception(requests.exceptions.HTTPError(error))

        except requests.exceptions.ConnectionError:
            msg = ("Cannot connect to PlexxiCore at " +
                   self.creds['auth_url'] + " with the username " +
                   self.creds['username'])
            raise Exception(requests.exceptions.ConnectionError(msg))

    def keystoneauth(self):
        """Acquire a keystone auth token for API calls

        Called when congress is running with keystone as the authentication
        method.This provides the driver a keystone token that is then placed
        in the header of API calls made to congress.
        """
        try:
            authreq = {
                "auth": {
                    "tenantName": self.creds['tenant_name'],
                    "passwordCredentials": {
                        "username": self.creds['keystone_user'],
                        "password": self.creds['keystone_pass']
                        }
                    }
                }
            headers = {'content-type': 'application/json',
                       'accept': 'application/json'}
            request = requests.post(url=self.keystone_url+'/v2.0/tokens',
                                    data=json.dumps(authreq),
                                    headers=headers)
            response = request.json()
            token = response['access']['token']['id']
            self.headers['X-Auth-Token'] = token
        except Exception:
            LOG.exception("Could not authenticate with keystone." +
                          "All automated API calls have been disabled")
            self.unique_names = False
            self.name_rule_needed = False

    def name_rule_check(self):
        """Checks to see if a RepeatedNames rule already exists

        This method is used to prevent the driver from recreating additional
        RepeatedNames tables each time congress is restarted.
        """
        try:
            table = requests.get(self.api_address + "/policies/" +
                                 "plexxi/rules",
                                 headers=self.headers)
            result = json.loads(table.text)
            for entry in result['results']:
                if entry['name'] == "RepeatedNames":
                    return False
            return True
        except Exception:
            LOG.exception("An error has occurred when accessing the " +
                          "Congress API.All automated API calls have been " +
                          "disabled.")
            self.unique_names = False
            self.name_rule_needed = False
            return False

    def name_rule_create(self):
        """Creates RepeatedName table for unique names policy.

        The RepeatedName table contains the name and plexxiUuid of
        VMs that have the same name in the Plexxi table and the Nova Table.
        """
        try:
            datasources = datasource_mgr.DataSourceManager.get_datasources()
            for datasource in datasources:
                if datasource['driver'] == 'nova':
                    repeated_name_rule = ('{"rule": "RepeatedName' +
                                          '(vname,pvuuid):-' + self.name +
                                          ':vms(0=pvuuid,1=vname),' +
                                          datasource['name'] +
                                          ':servers(1=vname)",' +
                                          '"name": "RepeatedNames"}')
                    requests.post(url=self.api_address +
                                  '/policies/plexxi/rules',
                                  data=repeated_name_rule,
                                  headers=self.headers)
                    self.name_rule_needed = False
                    break
        except Exception:
            LOG.exception("Could not create Repeated Name table")

    def name_response(self):
        """Checks for any entries in the RepeatedName table.

        For all entries found in the RepeatedName table, the corresponding
        VM will be then prefixed with 'conflict-' in PlexxiCore.
        """

        vmname = False
        vmuuid = False
        json_response = []
        self.name_cooldown = True
        try:
            plexxivms = VmwareVirtualMachine.getAll(session=self.exchange)
            table = requests.get(self.api_address + "/policies/" +
                                 "plexxi/tables/RepeatedName/rows",
                                 headers=self.headers)
            if table.text == "Authentication required":
                self.keystoneauth()
                table = requests.get(self.api_address + "/policies/" +
                                     "plexxi/tables/RepeatedName/rows",
                                     headers=self.headers)
            json_response = json.loads(table.text)

            for row in json_response['results']:
                vmname = row['data'][0]
                vmuuid = row['data'][1]
                if vmname and vmuuid:
                    for plexxivm in plexxivms:
                        if (plexxivm.getForeignUuid() == vmuuid):
                            new_vm_name = "Conflict-" + vmname
                            desc = ("Congress has found a VM with the same " +
                                    "name on the nova network. This vm " +
                                    "will now be renamed to " + new_vm_name)
                            job_name = (" Congress Driver:Changing virtual" +
                                        "machine, " + vmname + "\'s name")
                            changenamejob = Job.create(name=job_name,
                                                       description=desc + ".",
                                                       session=self.exchange)
                            changenamejob.begin()
                            plexxivm.setName(new_vm_name)
                            changenamejob.commit()
                            LOG.info(desc + " in PlexxiCore.")
        except Exception:
            LOG.exception("error in name_response")



# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_log import log as logging

from congress.datalog import compile
from congress.datalog import factset
from congress.datalog import utility


LOG = logging.getLogger(__name__)


class RuleSet(object):
    """RuleSet

    Keeps track of all rules for all tables.
    """

    # Internally:
    #  An index_name looks like this: (p, (2, 4)) which means this index is
    #  on table 'p' and it specifies columns 2 and 4.
    #
    #  An index_key looks like this: (p, (2, 'abc'), (4, 'def'))

    def __init__(self):
        self.rules = {}
        self.facts = {}

    def __str__(self):
        return str(self.rules) + " " + str(self.facts)

    def add_rule(self, key, rule):
        """Add a rule to the Ruleset

        @rule can be a Rule or a Fact. Returns True if add_rule() changes the
        RuleSet.
        """
        if isinstance(rule, compile.Fact):
            # If the rule is a Fact, then add it to self.facts.
            if key not in self.facts:
                self.facts[key] = factset.FactSet()
            return self.facts[key].add(rule)

        elif len(rule.body) == 0 and not rule.head.is_negated():
            # If the rule is a Rule, with no body, then it's a Fact, so
            # convert the Rule to a Fact to a Fact and add to self.facts.
            f = compile.Fact(key, (a.name for a in rule.head.arguments))
            if key not in self.facts:
                self.facts[key] = factset.FactSet()
            return self.facts[key].add(f)

        else:
            # else the rule is a regular rule, so add it to self.rules.
            if key in self.rules:
                return self.rules[key].add(rule)
            else:
                self.rules[key] = utility.OrderedSet([rule])
                return True

    def discard_rule(self, key, rule):
        """Remove a rule from the Ruleset

        @rule can be a Rule or a Fact. Returns True if discard_rule() changes
        the RuleSet.
        """
        if isinstance(rule, compile.Fact):
            # rule is a Fact, so remove from self.facts
            if key in self.facts:
                changed = self.facts[key].remove(rule)
                if len(self.facts[key]) == 0:
                    del self.facts[key]
                return changed
            return False

        elif not len(rule.body):
            # rule is a Rule, but without a body so it will be in self.facts.
            if key in self.facts:
                fact = compile.Fact(key, [a.name for a in rule.head.arguments])
                changed = self.facts[key].remove(fact)
                if len(self.facts[key]) == 0:
                    del self.facts[key]
                return changed
            return False

        else:
            # rule is a Rule with a body, so remove from self.rules.
            if key in self.rules:
                changed = self.rules[key].discard(rule)
                if len(self.rules[key]) == 0:
                    del self.rules[key]
                return changed
            return False

    def keys(self):
        return list(self.facts.keys()) + list(self.rules.keys())

    def __contains__(self, key):
        return key in self.facts or key in self.rules

    def contains(self, key, rule):
        if isinstance(rule, compile.Fact):
            return key in self.facts and rule in self.facts[key]
        elif isinstance(rule, compile.Literal):
            if key not in self.facts:
                return False
            fact = compile.Fact(key, [a.name for a in rule.arguments])
            return fact in self.facts[key]
        elif not len(rule.body):
            if key not in self.facts:
                return False
            fact = compile.Fact(key, [a.name for a in rule.head.arguments])
            return fact in self.facts[key]
        else:
            return key in self.rules and rule in self.rules[key]

    def get_rules(self, key, match_literal=None):
        facts = []

        if (match_literal and not match_literal.is_negated() and
                key in self.facts):
            # If the caller supplies a literal to match against, then use an
            # index to find the matching rules.
            bound_arguments = tuple([i for i, arg
                                     in enumerate(match_literal.arguments)
                                     if not arg.is_variable()])
            if (bound_arguments and
                    not self.facts[key].has_index(bound_arguments)):
                # The index does not exist, so create it.
                self.facts[key].create_index(bound_arguments)

            partial_fact = tuple(
                [(i, arg.name)
                 for i, arg in enumerate(match_literal.arguments)
                 if not arg.is_variable()])
            facts = list(self.facts[key].find(partial_fact))
        else:
            # There is no usable match_literal, so get all facts for the
            # table.
            facts = list(self.facts.get(key, ()))

        # Convert native tuples to Rule objects.

        # TODO(alex): This is inefficient because it creates Literal and Rule
        # objects.  It would be more efficient to change the TopDownTheory and
        # unifier to handle Facts natively.
        fact_rules = []
        for fact in facts:
            # Setting use_modules=False so we don't split up tablenames.
            #   This allows us to choose at compile-time whether to split
            #   the tablename up.
            literal = compile.Literal(
                key, [compile.Term.create_from_python(x) for x in fact],
                use_modules=False)
            fact_rules.append(compile.Rule(literal, ()))

        return fact_rules + list(self.rules.get(key, ()))

    def clear(self):
        self.rules = {}
        self.facts = {}

    def clear_table(self, table):
        self.rules[table] = utility.OrderedSet()
        self.facts[table] = factset.FactSet()



# Copyright (c) 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import json

from oslo_log import log as logging
import sqlalchemy as sa
from sqlalchemy.orm import exc as db_exc

from congress.db import api as db
from congress.db import model_base


LOG = logging.getLogger(__name__)


class Datasource(model_base.BASE, model_base.HasId):
    __tablename__ = 'datasources'

    name = sa.Column(sa.String(255), unique=True)
    driver = sa.Column(sa.String(255))
    config = sa.Column(sa.Text(), nullable=False)
    description = sa.Column(sa.Text(), nullable=True)
    enabled = sa.Column(sa.Boolean, default=True)

    def __init__(self, id_, name, driver, config, description,
                 enabled=True):
        self.id = id_
        self.name = name
        self.driver = driver
        self.config = json.dumps(config)
        self.description = description
        self.enabled = enabled


def add_datasource(id_, name, driver, config, description,
                   enabled, session=None):
    session = session or db.get_session()
    with session.begin(subtransactions=True):
        datasource = Datasource(
            id_=id_,
            name=name,
            driver=driver,
            config=config,
            description=description,
            enabled=enabled)
        session.add(datasource)
    return datasource


def delete_datasource(id_, session=None):
    session = session or db.get_session()
    return session.query(Datasource).filter(
        Datasource.id == id_).delete()


def get_datasource(id_, session=None):
    session = session or db.get_session()
    try:
        return (session.query(Datasource).
                filter(Datasource.id == id_).
                one())
    except db_exc.NoResultFound:
        pass


def get_datasource_by_name(name, session=None, deleted=False):
    session = session or db.get_session()
    try:
        return (session.query(Datasource).
                filter(Datasource.name == name).
                one())
    except db_exc.NoResultFound:
        pass


def get_datasources(session=None, deleted=False):
    session = session or db.get_session()
    return (session.query(Datasource).
            all())



# Copyright (c) 2013 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

try:
    # For Python 3
    import http.client as httplib
except ImportError:
    import httplib
import json
import re
import uuid

from oslo_config import cfg
from oslo_log import log as logging
import six
import webob
import webob.dec

from congress.api import error_codes
from congress.common import policy
from congress import exception


LOG = logging.getLogger(__name__)


def error_response(status, error_code, description, data=None):
    """Construct and return an error response.

    Args:
        status: The HTTP status code of the response.
        error_code: The application-specific error code.
        description: Friendly G11N-enabled string corresponding to error_code.
        data: Additional data (not G11N-enabled) for the API consumer.
    """
    raw_body = {'error': {
        'message': description,
        'error_code': error_code,
        'error_data': data
        }
    }
    body = '%s\n' % json.dumps(raw_body)
    return webob.Response(body=body, status=status,
                          content_type='application/json')


NOT_FOUND_RESPONSE = error_response(httplib.NOT_FOUND,
                                    httplib.NOT_FOUND,
                                    "The resouce could not be found.")
NOT_SUPPORTED_RESPONSE = error_response(httplib.NOT_IMPLEMENTED,
                                        httplib.NOT_IMPLEMENTED,
                                        "Method not supported")
INTERNAL_ERROR_RESPONSE = error_response(httplib.INTERNAL_SERVER_ERROR,
                                         httplib.INTERNAL_SERVER_ERROR,
                                         "Internal server error")


class DataModelException(Exception):
    """Congress API Data Model Exception

    Custom exception raised by API Data Model methods to communicate errors to
    the API framework.
    """

    def __init__(self, error_code, description, data=None,
                 http_status_code=httplib.BAD_REQUEST):
        super(DataModelException, self).__init__(description)
        self.error_code = error_code
        self.description = description
        self.data = data
        self.http_status_code = http_status_code

    @classmethod
    def create(cls, error):
        """Generate a DataModelException from an existing CongressException.

        :param error has a 'name' field corresponding to an error_codes
        error-name.  It may also have a 'data' field.
        Returns a DataModelException properly populated.
        """
        name = getattr(error, "name", None)
        if name:
            error_code = error_codes.get_num(name)
            description = error_codes.get_desc(name)
            http_status_code = error_codes.get_http(name)
        else:
            # Check if it's default http error or else return 'Unknown error'
            error_code = error.code or httplib.BAD_REQUEST
            if error_code not in httplib.responses:
                error_code = httplib.BAD_REQUEST
            description = httplib.responses.get(error_code, "Unknown error")
            http_status_code = error_code

        if str(error):
            description += "::" + str(error)
        return cls(error_code=error_code,
                   description=description,
                   data=getattr(error, 'data', None),
                   http_status_code=http_status_code)

    def rest_response(self):
        return error_response(self.http_status_code, self.error_code,
                              self.description, self.data)


class AbstractApiHandler(object):
    """Abstract handler for API requests.

    Attributes:
        path_regex: The regular expression matching paths supported by this
            handler.
    """

    def __init__(self, path_regex):
        if path_regex[-1] != '$':
            path_regex += "$"
        # we only use 'match' so no need to mark the beginning of string
        self.path_regex = path_regex
        self.path_re = re.compile(path_regex)

    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, self.path_re.pattern)

    def _get_context(self, request):
        """Return dict of variables in request path."""
        m = self.path_re.match(request.path)
        # remove all the None values before returning
        return dict([(k, v) for k, v in m.groupdict().items()
                     if v is not None])

    def _parse_json_body(self, request):
        content_type = (request.content_type or "application/json").lower()
        if content_type != 'application/json':
            raise DataModelException(
                400, "Unsupported Content-Type; must be 'application/json'")
        if request.charset != 'UTF-8':
            raise DataModelException(
                400, "Unsupported charset: must be 'UTF-8'")
        try:
            request.parsed_body = json.loads(request.body.decode('utf-8'))
        except ValueError as e:
            msg = "Failed to parse body as %s: %s" % (content_type, e)
            raise DataModelException(400, msg)
        return request.parsed_body

    def handles_request(self, request):
        """Return true iff handler supports the request."""
        m = self.path_re.match(request.path)
        return m is not None

    def handle_request(self, request):
        """Handle a REST request.

        Args:
           request: A webob request object.

        Returns:
            A webob response object.
        """
        return NOT_SUPPORTED_RESPONSE


class ElementHandler(AbstractApiHandler):
    """API handler for REST element resources.

    REST elements represent individual entities in the data model, and often
    support the following operations:
        - Read a representation of the element
        - Update (replace) the entire element with a new version
        - Update (patch) parts of the element with new values
        - Delete the element

    Elements may also exhibit 'controller' semantics for RPC-style method
    invocation, however this is not currently supported.
    """

    def __init__(self, path_regex, model,
                 collection_handler=None, allow_read=True, allow_actions=True,
                 allow_replace=True, allow_update=True, allow_delete=True):
        """Initialize an element handler.

        Args:
            path_regex: A regular expression that matches the full path
                to the element.  If multiple handlers match a request path,
                the handler with the highest registration search_index wins.
            model: A resource data model instance
            collection_handler: The collection handler this element
                is a member of or None if the element is not a member of a
                collection.  (Used for named creation of elements)
            allow_read: True if element supports read
            allow_replace: True if element supports replace
            allow_update: True if element supports update
            allow_delete: True if element supports delete

        """
        super(ElementHandler, self).__init__(path_regex)
        self.model = model
        self.collection_handler = collection_handler
        self.allow_read = allow_read
        self.allow_actions = allow_actions
        self.allow_replace = allow_replace
        self.allow_update = allow_update
        self.allow_delete = allow_delete

    def _get_element_id(self, request):
        m = self.path_re.match(request.path)
        if m.groups():
            return m.groups()[-1]  # TODO(pballand): make robust
        return None

    def handle_request(self, request):
        """Handle a REST request.

        Args:
           request: A webob request object.

        Returns:
            A webob response object.
        """
        if request.method == 'GET' and self.allow_read:
            return self.read(request)
        elif request.method == 'POST' and self.allow_actions:
            return self.action(request)
        elif request.method == 'PUT' and self.allow_replace:
            return self.replace(request)
        elif request.method == 'PATCH' and self.allow_update:
            return self.update(request)
        elif request.method == 'DELETE' and self.allow_delete:
            return self.delete(request)
        return NOT_SUPPORTED_RESPONSE

    def read(self, request):
        if not hasattr(self.model, 'get_item'):
            return NOT_SUPPORTED_RESPONSE

        id_ = self._get_element_id(request)
        item = self.model.get_item(id_, request.params,
                                   context=self._get_context(request))
        if item is None:
            return error_response(httplib.NOT_FOUND, 404, 'Not found')
        return webob.Response(body="%s\n" % json.dumps(item),
                              status=httplib.OK,
                              content_type='application/json')

    def action(self, request):
        # Non-CRUD operations must specify an 'action' parameter
        action = request.params.getall('action')
        if len(action) != 1:
            if len(action) > 1:
                errstr = "Action parameter may not be provided multiple times."
            else:
                errstr = "Missing required action parameter."
            return error_response(httplib.BAD_REQUEST, 400, errstr)
        model_method = "%s_action" % action[0].replace('-', '_')
        f = getattr(self.model, model_method, None)
        if f is None:
            return NOT_SUPPORTED_RESPONSE
        try:
            response = f(request.params, context=self._get_context(request),
                         request=request)
            if isinstance(response, webob.Response):
                return response
            return webob.Response(body="%s\n" % json.dumps(response),
                                  status=httplib.OK,
                                  content_type='application/json')
        except TypeError:
            LOG.exception("Error occurred")
            return NOT_SUPPORTED_RESPONSE

    def replace(self, request):
        if not hasattr(self.model, 'update_item'):
            return NOT_SUPPORTED_RESPONSE

        id_ = self._get_element_id(request)
        try:
            item = self._parse_json_body(request)
            self.model.update_item(id_, item, request.params,
                                   context=self._get_context(request))
        except KeyError as e:
            if (self.collection_handler and
                    getattr(self.collection_handler, 'allow_named_create',
                            False)):
                return self.collection_handler.create_member(request, id_=id_)
            return error_response(httplib.NOT_FOUND, 404,
                                  e.message or 'Not found')
        return webob.Response(body="%s\n" % json.dumps(item),
                              status=httplib.OK,
                              content_type='application/json')

    def update(self, request):
        if not (hasattr(self.model, 'update_item') or
                hasattr(self.model, 'get_item')):
            return NOT_SUPPORTED_RESPONSE

        context = self._get_context(request)
        id_ = self._get_element_id(request)
        item = self.model.get_item(id_, request.params, context=context)
        if item is None:
            return error_response(httplib.NOT_FOUND, 404, 'Not found')

        updates = self._parse_json_body(request)
        item.update(updates)
        self.model.update_item(id_, item, request.params, context=context)
        return webob.Response(body="%s\n" % json.dumps(item),
                              status=httplib.OK,
                              content_type='application/json')

    def delete(self, request):
        if not hasattr(self.model, 'delete_item'):
            return NOT_SUPPORTED_RESPONSE

        id_ = self._get_element_id(request)
        try:
            item = self.model.delete_item(
                id_, request.params, context=self._get_context(request))
            return webob.Response(body="%s\n" % json.dumps(item),
                                  status=httplib.OK,
                                  content_type='application/json')
        except KeyError as e:
            LOG.exception("Error occurred")
            return error_response(httplib.NOT_FOUND, 404,
                                  e.message or 'Not found')


class CollectionHandler(AbstractApiHandler):
    """API handler for REST collection resources.

    REST collections represent collections of entities in the data model, and
    often support the following operations:
        - List elements in the collection
        - Create new element in the collection

    The following less-common collection operations are NOT SUPPORTED:
        - Replace all elements in the collection
        - Delete all elements in the collection
    """

    def __init__(self, path_regex, model,
                 allow_named_create=True, allow_list=True, allow_create=True,
                 allow_update=False):
        """Initialize a collection handler.

        Args:
            path_regex: A regular expression matching the collection base path.
            model: A resource data model instance
            allow_named_create: True if caller can specify ID of new items.
            allow_list: True if collection supports listing elements.
            allow_create: True if collection supports creating elements.
        """
        super(CollectionHandler, self).__init__(path_regex)
        self.model = model
        self.allow_named_create = allow_named_create
        self.allow_list = allow_list
        self.allow_create = allow_create
        self.allow_update = allow_update

    def handle_request(self, request):
        """Handle a REST request.

        Args:
           request: A webob request object.

        Returns:
            A webob response object.
        """
        # NOTE(arosen): only do policy.json if keystone is used for now.
        if cfg.CONF.auth_strategy == "keystone":
            context = request.environ['congress.context']
            target = {
                'project_id': context.project_id,
                'user_id': context.user_id
            }
            # NOTE(arosen): today congress only enforces API policy on which
            # API calls we allow tenants to make with their given roles.
            action_type = self._get_action_type(request.method)
            # FIXME(arosen): There should be a cleaner way to do this.
            model_name = self.path_regex.split('/')[1]
            action = "%s_%s" % (action_type, model_name)
            # TODO(arosen): we should handle serializing the
            # response in one place
            try:
                policy.enforce(context, action, target)
            except exception.PolicyNotAuthorized as e:
                LOG.info(e)
                return webob.Response(body=six.text_type(e), status=e.code,
                                      content_type='application/json')
        if request.method == 'GET' and self.allow_list:
            return self.list_members(request)
        elif request.method == 'POST' and self.allow_create:
            return self.create_member(request)
        elif request.method == 'PUT' and self.allow_update:
            return self.update_members(request)
        return NOT_SUPPORTED_RESPONSE

    def _get_action_type(self, method):
        if method == 'GET':
            return 'get'
        elif method == 'POST':
            return 'create'
        elif method == 'DELETE':
            return 'delete'
        elif method == 'PUT' or method == 'PATCH':
            return 'update'
        else:
            # should never get here but just in case ;)
            # FIXME(arosen) raise NotImplemented instead and
            # make sure we return that as an http code.
            raise TypeError("Invalid HTTP Method")

    def list_members(self, request):
        if not hasattr(self.model, 'get_items'):
            return NOT_SUPPORTED_RESPONSE
        items = self.model.get_items(request.params,
                                     context=self._get_context(request))
        if items is None:
            return error_response(httplib.NOT_FOUND, 404, 'Not found')
        elif 'results' not in items:
            return error_response(httplib.NOT_FOUND, 404, 'Not found')

        body = "%s\n" % json.dumps(items, indent=2)
        return webob.Response(body=body, status=httplib.OK,
                              content_type='application/json')

    def create_member(self, request, id_=None):
        if not hasattr(self.model, 'add_item'):
            return NOT_SUPPORTED_RESPONSE
        item = self._parse_json_body(request)
        context = self._get_context(request)
        try:
            id_, item = self.model.add_item(
                item, request.params, id_, context=context)
        except KeyError as e:
            LOG.exception("Error occurred")
            return error_response(httplib.CONFLICT, httplib.CONFLICT,
                                  e.message or 'Element already exists')
        item['id'] = id_

        return webob.Response(body="%s\n" % json.dumps(item),
                              status=httplib.CREATED,
                              content_type='application/json',
                              location="%s/%s" % (request.path, id_))

    def update_members(self, request):
        if not hasattr(self.model, 'update_items'):
            return NOT_SUPPORTED_RESPONSE
        items = self._parse_json_body(request)
        context = self._get_context(request)
        try:
            self.model.update_items(items, request.params, context)
        except KeyError as e:
            LOG.exception("Error occured")
            return error_response(httplib.BAD_REQUEST, httplib.BAD_REQUEST,
                                  e.message or
                                  'Update %s Failed' % context['table_id'])
        return webob.Response(body="", status=httplib.OK,
                              content_type='application/json')


class SimpleDataModel(object):
    """A container providing access to a single type of data."""

    def __init__(self, model_name):
        self.model_name = model_name
        self.items = {}

    @staticmethod
    def _context_str(context):
        context = context or {}
        return ".".join(
            ["%s:%s" % (k, context[k]) for k in sorted(context.keys())])

    def get_items(self, params, context=None):
        """Get items in model.

        Args:
            params: A dict-like object containing parameters
                    from the request query string and body.
            context: Key-values providing frame of reference of request

        Returns: A dict containing at least a 'results' key whose value is
                 a list of items in the model.  Additional keys set in the
                 dict will also be rendered for the user.
        """
        cstr = self._context_str(context)
        results = list(self.items.setdefault(cstr, {}).values())
        return {'results': results}

    def add_item(self, item, params, id_=None, context=None):
        """Add item to model.

        Args:
            item: The item to add to the model
            params: A dict-like object containing parameters
                    from the request query string and body.
            id_: The ID of the item, or None if an ID should be generated
            context: Key-values providing frame of reference of request

        Returns:
             Tuple of (ID, newly_created_item)

        Raises:
            KeyError: ID already exists.
            DataModelException: Addition cannot be performed.
        """
        cstr = self._context_str(context)
        if id_ is None:
            id_ = str(uuid.uuid4())
        if id_ in self.items.setdefault(cstr, {}):
            raise KeyError("Cannot create item with ID '%s': "
                           "ID already exists" % id_)
        self.items[cstr][id_] = item
        return (id_, item)

    def get_item(self, id_, params, context=None):
        """Retrieve item with id id_ from model.

        Args:
            id_: The ID of the item to retrieve
            params: A dict-like object containing parameters
                    from the request query string and body.
            context: Key-values providing frame of reference of request

        Returns:
             The matching item or None if item with id_ does not exist.
        """
        cstr = self._context_str(context)
        return self.items.setdefault(cstr, {}).get(id_)

    def update_item(self, id_, item, params, context=None):
        """Update item with id_ with new data.

        Args:
            id_: The ID of the item to be updated
            item: The new item
            params: A dict-like object containing parameters
                    from the request query string and body.
            context: Key-values providing frame of reference of request

        Returns:
             The updated item.

        Raises:
            KeyError: Item with specified id_ not present.
            DataModelException: Update cannot be performed.
        """
        cstr = self._context_str(context)
        if id_ not in self.items.setdefault(cstr, {}):
            raise KeyError("Cannot update item with ID '%s': "
                           "ID does not exist" % id_)
        self.items.setdefault(cstr, {})[id_] = item
        return item

    def delete_item(self, id_, params, context=None):
        """Remove item from model.

        Args:
            id_: The ID of the item to be removed
            params: A dict-like object containing parameters
                    from the request query string and body.
            context: Key-values providing frame of reference of request

        Returns:
             The removed item.

        Raises:
            KeyError: Item with specified id_ not present.
        """
        cstr = self._context_str(context)
        ret = self.items.setdefault(cstr, {})[id_]
        del self.items[cstr][id_]
        return ret

    def update_items(self, items, params, context=None):
        """Update items in the model.

        Args:
            items: A dict-like object containing new data
            params: A dict-like object containing parameters
            context: Key-values providing frame of reference of request
        Returns:
            None.
        """
        self.items = items



# Copyright (c) 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Policy Engine For Auth on API calls."""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_config import cfg
from oslo_policy import policy

from congress import exception


_ENFORCER = None


def reset():
    global _ENFORCER
    if _ENFORCER:
        _ENFORCER.clear()
        _ENFORCER = None


def init(policy_file=None, rules=None, default_rule=None, use_conf=True):
    """Init an Enforcer class.

       :param policy_file: Custom policy file to use, if none is specified,
                           `CONF.policy_file` will be used.
       :param rules: Default dictionary / Rules to use. It will be
                     considered just in the first instantiation.
       :param default_rule: Default rule to use, CONF.default_rule will
                            be used if none is specified.
       :param use_conf: Whether to load rules from config file.
    """
    global _ENFORCER
    if not _ENFORCER:
        _ENFORCER = policy.Enforcer(cfg.CONF, policy_file=policy_file,
                                    rules=rules,
                                    default_rule=default_rule,
                                    use_conf=use_conf)


def set_rules(rules, overwrite=True, use_conf=False):
    """Set rules based on the provided dict of rules.

       :param rules: New rules to use. It should be an instance of dict.
       :param overwrite: Whether to overwrite current rules or update them
                         with the new rules.
       :param use_conf: Whether to reload rules from config file.
    """
    init(use_conf=False)
    _ENFORCER.set_rules(rules, overwrite, use_conf)


def enforce(context, action, target, do_raise=True, exc=None):
    """Verifies that the action is valid on the target in this context.

       :param context: congress context
       :param action: string representing the action to be checked
           this should be colon separated for clarity.
           i.e. ``compute:create_instance``,
           ``compute:attach_volume``,
           ``volume:attach_volume``
       :param target: dictionary representing the object of the action
           for object creation this should be a dictionary representing the
           location of the object e.g. ``{'project_id': context.project_id}``
       :param do_raise: if True (the default), raises PolicyNotAuthorized;
           if False, returns False

       :raises congress.exception.PolicyNotAuthorized: if verification fails
           and do_raise is True.

       :return: returns a non-False value (not necessarily "True") if
           authorized, and the exact value False if not authorized and
           do_raise is False.
    """
    init()
    credentials = context.to_dict()
    if not exc:
        exc = exception.PolicyNotAuthorized
    return _ENFORCER.enforce(action, target, credentials, do_raise=do_raise,
                             exc=exc, action=action)


def check_is_admin(context):
    """Whether or not roles contains 'admin' role according to policy setting.

    """
    init()
    # the target is user-self
    credentials = context.to_dict()
    target = credentials
    return _ENFORCER.enforce('context_is_admin', target, credentials)


@policy.register('is_admin')
class IsAdminCheck(policy.Check):
    """An explicit check for is_admin."""

    def __init__(self, kind, match):
        """Initialize the check."""

        self.expected = (match.lower() == 'true')

        super(IsAdminCheck, self).__init__(kind, str(self.expected))

    def __call__(self, target, creds, enforcer):
        """Determine whether is_admin matches the requested value."""

        return creds['is_admin'] == self.expected


def get_rules():
    if _ENFORCER:
        return _ENFORCER.rules



# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_utils import uuidutils

from congress.db import db_policy_rules
from congress.tests import base


class TestPolicyRulesDb(base.SqlTestCase):

    def test_add_policy_rule(self):
        id = uuidutils.generate_uuid()
        rule_str = "p(x) :- q(x)"
        policy_name = "classification"
        comment = "None"
        rule = db_policy_rules.add_policy_rule(id=id,
                                               policy_name=policy_name,
                                               rule=rule_str,
                                               comment=comment)
        self.assertEqual(id, rule.id)
        self.assertEqual(policy_name, rule.policy_name)
        self.assertEqual(rule_str, rule.rule)
        self.assertEqual(comment, rule.comment)

    def test_add_policy_rule_with_name(self):
        id = uuidutils.generate_uuid()
        rule_str = "p(x) :- q(x)"
        policy_name = "classification"
        comment = "None"
        rule_name = "classification_rule"
        rule = db_policy_rules.add_policy_rule(id=id,
                                               policy_name=policy_name,
                                               rule=rule_str,
                                               comment=comment,
                                               rule_name=rule_name)
        self.assertEqual(id, rule.id)
        self.assertEqual(policy_name, rule.policy_name)
        self.assertEqual(rule_str, rule.rule)
        self.assertEqual(comment, rule.comment)
        self.assertEqual(rule_name, rule.name)

    def test_add_get_policy_rule(self):
        id = uuidutils.generate_uuid()
        rule_str = "p(x) :- q(x)"
        policy_name = "classification"
        comment = "None"
        db_policy_rules.add_policy_rule(id=id,
                                        policy_name=policy_name,
                                        rule=rule_str,
                                        comment=comment)
        rule = db_policy_rules.get_policy_rule(id, policy_name)
        self.assertEqual(id, rule.id)
        self.assertEqual(policy_name, rule.policy_name)
        self.assertEqual(rule_str, rule.rule)
        self.assertEqual(comment, rule.comment)

    def test_add_delete_get_policy_rule(self):
        id = uuidutils.generate_uuid()
        rule_str = "p(x) :- q(x)"
        policy_name = "classification"
        comment = "None"
        db_policy_rules.add_policy_rule(id=id,
                                        policy_name=policy_name,
                                        rule=rule_str,
                                        comment=comment)
        db_policy_rules.delete_policy_rule(id)
        rule = db_policy_rules.get_policy_rule(id, policy_name)
        self.assertIsNone(rule)

    def test_add_delete_get_deleted_policy_rule(self):
        id = uuidutils.generate_uuid()
        rule_str = "p(x) :- q(x)"
        policy_name = "classification"
        comment = "None"
        rule1 = db_policy_rules.add_policy_rule(id=id,
                                                policy_name=policy_name,
                                                rule=rule_str,
                                                comment=comment)
        db_policy_rules.delete_policy_rule(id)
        rule2 = db_policy_rules.get_policy_rule(id, policy_name, deleted=True)
        self.assertEqual(rule1.id, rule2.id)
        self.assertNotEqual(rule1.deleted, rule2.deleted)

    def test_add_two_rules_and_get(self):
        id1 = uuidutils.generate_uuid()
        rule1_str = "p(x) :- q(x)"
        id2 = uuidutils.generate_uuid()
        rule2_str = "z(x) :- q(x)"
        policy_name = "classification"
        comment = "None"
        db_policy_rules.add_policy_rule(id=id1,
                                        policy_name=policy_name,
                                        rule=rule1_str,
                                        comment=comment)

        db_policy_rules.add_policy_rule(id=id2,
                                        policy_name=policy_name,
                                        rule=rule2_str,
                                        comment=comment)

        rules = db_policy_rules.get_policy_rules(policy_name)
        self.assertEqual(len(rules), 2)
        self.assertEqual(id1, rules[0].id)
        self.assertEqual(policy_name, rules[0].policy_name)
        self.assertEqual(rule1_str, rules[0].rule)
        self.assertEqual(comment, rules[0].comment)
        self.assertEqual(id2, rules[1].id)
        self.assertEqual(policy_name, rules[1].policy_name)
        self.assertEqual(rule2_str, rules[1].rule)
        self.assertEqual(comment, rules[1].comment)
        self.assertEqual(len(db_policy_rules.get_policy_rules()), 2)

    def test_is_soft_deleted_not_deleted(self):
        uuid = uuidutils.generate_uuid()
        self.assertEqual('', db_policy_rules.is_soft_deleted(uuid, False))

    def test_is_soft_deleted_is_deleted(self):
        uuid = uuidutils.generate_uuid()
        self.assertEqual(uuid, db_policy_rules.is_soft_deleted(uuid, True))



# Copyright 2014 VMware.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from django.core.urlresolvers import reverse
from django.template.defaultfilters import unordered_list
from django.utils.translation import ugettext_lazy as _
from horizon import tables


def get_resource_url(obj):
    return reverse('horizon:admin:datasources:datasource_table_detail',
                   args=(obj['datasource_id'], obj['table_id']))


class DataSourcesTablesTable(tables.DataTable):
    name = tables.Column("name", verbose_name=_("Table Name"),
                         link=get_resource_url)
    datasource_name = tables.Column("datasource_name",
                                    verbose_name=_("Service"))
    datasource_driver = tables.Column("datasource_driver",
                                      verbose_name=_("Driver"))

    class Meta:
        name = "datasources_tables"
        verbose_name = _("Service Data")
        hidden_title = False


def get_policy_link(datum):
    return reverse('horizon:admin:policies:detail',
                   args=(datum['policy_name'],))


def get_policy_table_link(datum):
    return reverse('horizon:admin:datasources:policy_table_detail',
                   args=(datum['policy_name'], datum['name']))


class PoliciesTablesTable(tables.DataTable):
    name = tables.Column("name", verbose_name=_("Table Name"),
                         link=get_policy_table_link)
    policy_name = tables.Column("policy_name", verbose_name=_("Policy"),
                                link=get_policy_link)
    policy_owner_id = tables.Column("policy_owner_id",
                                    verbose_name=_("Owner ID"))

    class Meta:
        name = "policies_tables"
        verbose_name = _("Policy Data")
        hidden_title = False


class DataSourceRowsTable(tables.DataTable):
    class Meta:
        name = "datasource_rows"
        verbose_name = _("Rows")
        hidden_title = False


class DataSourceStatusesTable(tables.DataTable):
    datasource_name = tables.Column("service",
                                    verbose_name=_("Service"))
    last_updated = tables.Column("last_updated",
                                 verbose_name=_("Last Updated"))
    subscriptions = tables.Column("subscriptions",
                                  verbose_name=_("Subscriptions"),
                                  wrap_list=True, filters=(unordered_list,))
    last_error = tables.Column("last_error", verbose_name=_("Last Error"))
    subscribers = tables.Column("subscribers", verbose_name=_("Subscribers"),
                                wrap_list=True, filters=(unordered_list,))
    initialized = tables.Column("initialized", verbose_name=_("Initialized"))
    number_of_updates = tables.Column("number_of_updates",
                                      verbose_name=_("Number of Updates"))

    class Meta:
        name = "service_status"
        verbose_name = _("Service Status")
        hidden_title = False



# Copyright (c) 2015 NTT All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_config import cfg
from oslo_log import log as logging

from congress.api import webservice

LOG = logging.getLogger(__name__)


def create_table_dict(tablename, schema):
    cols = [{'name': x['name'], 'description': x['desc']}
            if isinstance(x, dict)
            else {'name': x, 'description': 'None'}
            for x in schema[tablename]]
    return {'table_id': tablename,
            'columns': cols}


def get_id_from_context(context, datasource_mgr=None, policy_engine=None):
    if cfg.CONF.distributed_architecture:
        datasource_mgr = context.get('ds_id')

    if 'ds_id' in context:
        return datasource_mgr, context.get('ds_id')
    elif 'policy_id' in context:
        return policy_engine, context.get('policy_id')
    else:
        msg = "Internal error: context %s should have included " % str(context)
        "either ds_id or policy_id"
        try:  # Py3: ensure LOG.exception is inside except
            raise webservice.DataModelException('404', msg)
        except webservice.DataModelException:
            LOG.exception(msg)
            raise



# Copyright 2015 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import mock
from oslo_config import cfg

from congress.api import rule_model
from congress.api import webservice
from congress.datalog import base as datalogbase
from congress.datalog import compile
from congress import harness
from congress.tests import base
from congress.tests import helper


class TestRuleModel(base.SqlTestCase):
    def setUp(self):
        super(TestRuleModel, self).setUp()
        # Here we load the fake driver
        cfg.CONF.set_override(
            'drivers',
            ['congress.tests.fake_datasource.FakeDataSource'])

        self.cage = harness.create(helper.root_path())
        self.engine = self.cage.service_object('engine')
        self.rule_model = rule_model.RuleModel("rule_model", {},
                                               policy_engine=self.engine)
        self.context = {'policy_id': 'action'}
        self._add_test_rule()

    def tearDown(self):
        super(TestRuleModel, self).tearDown()

    def _add_test_rule(self):
        test_rule1 = {
            "rule": "p(x) :- q(x)",
            "name": "test-rule1",
            "comment": "test-comment"
        }
        test_rule2 = {
            "rule": 'p(x) :- q(x), not r(x)',
            "name": "test-rule2",
            "comment": "test-comment-2"
        }
        test_rule_id, obj = self.rule_model.add_item(test_rule1, {},
                                                     context=self.context)
        test_rule1["id"] = test_rule_id
        self.rule1 = test_rule1

        test_rule_id, obj = self.rule_model.add_item(test_rule2, {},
                                                     context=self.context)
        test_rule2["id"] = test_rule_id
        self.rule2 = test_rule2

    @mock.patch.object(rule_model.RuleModel, 'policy_name')
    def test_add_rule_with_invalid_policy(self, policy_name_mock):
        test_rule = {'rule': 'p()', 'name': 'test'}
        policy_name_mock.return_value = 'invalid'
        self.assertRaises(webservice.DataModelException,
                          self.rule_model.add_item,
                          test_rule, {})

    def test_add_rule_with_colrefs(self):
        engine = self.engine
        engine.create_policy('beta', kind=datalogbase.DATASOURCE_POLICY_TYPE)
        engine.set_schema(
            'beta', compile.Schema({'q': ("name", "status", "year")}))
        # insert/retrieve rule with column references
        # just testing that no errors are thrown--correctness tested elsewhere
        # Assuming that api-models are pass-throughs to functionality
        (id1, _) = self.rule_model.add_item(
            {'rule': 'p(x) :- beta:q(name=x)'},
            {}, context=self.context)
        self.rule_model.get_item(id1, {}, context=self.context)

    def test_add_rule_with_bad_colrefs(self):
        engine = self.engine
        engine.create_policy('beta')   # not datasource policy
        # exception because col refs over non-datasource policy
        self.assertRaises(
            webservice.DataModelException,
            self.rule_model.add_item,
            {'rule': 'p(x) :- beta:q(name=x)'},
            {}, context=self.context)

    def test_get_items(self):
        ret = self.rule_model.get_items({}, context=self.context)
        self.assertTrue(all(p in ret['results']
                            for p in [self.rule1, self.rule2]))

    def test_get_item(self):
        expected_ret = self.rule1
        ret = self.rule_model.get_item(self.rule1["id"], {},
                                       context=self.context)
        self.assertEqual(expected_ret, ret)

    def test_get_invalid_item(self):
        expected_ret = None
        ret = self.rule_model.get_item('invalid-id', {}, context=self.context)
        self.assertEqual(expected_ret, ret)

    def test_delete_item(self):
        expected_ret = self.rule1

        ret = self.rule_model.delete_item(self.rule1['id'], {},
                                          context=self.context)
        self.assertEqual(expected_ret, ret)

        expected_ret = None
        ret = self.rule_model.get_item(self.rule1['id'], {},
                                       context=self.context)
        self.assertEqual(expected_ret, ret)

    def test_rule_api_model_errors(self):
        """Test syntax errors.

        Test that syntax errors thrown by the policy runtime
        are returned properly to the user so they can see the
        error messages.
        """
        # lexer error
        with self.assertRaisesRegexp(
                webservice.DataModelException,
                "Lex failure"):
            self.rule_model.add_item({'rule': 'p#'}, {}, context=self.context)

        # parser error
        with self.assertRaisesRegexp(
                webservice.DataModelException,
                "Parse failure"):
            self.rule_model.add_item({'rule': 'p('}, {}, context=self.context)

        # single-rule error: safety in the head
        with self.assertRaisesRegexp(
                webservice.DataModelException,
                "Variable x found in head but not in body"):
            # TODO(ramineni):check for action
            self.context = {'policy_id': 'classification'}
            self.rule_model.add_item({'rule': 'p(x,y) :- q(y)'}, {},
                                     context=self.context)

        # multi-rule error: recursion through negation
        self.rule_model.add_item({'rule': 'p(x) :- q(x), not r(x)'}, {},
                                 context=self.context)
        with self.assertRaisesRegexp(
                webservice.DataModelException,
                "Rules are recursive"):
            self.rule_model.add_item({'rule': 'r(x) :- q(x), not p(x)'}, {},
                                     context=self.context)

        self.rule_model.add_item({'rule': 'p(x) :- q(x)'}, {},
                                 context=self.context)
        # duplicate rules
        with self.assertRaisesRegexp(
                webservice.DataModelException,
                "Rule already exists"):
            self.rule_model.add_item({'rule': 'p(x) :- q(x)'}, {},
                                     context=self.context)
